# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Unit test for the extensions.diadefslib modules."""

# pylint: disable=redefined-outer-name

from __future__ import annotations

import sys
from collections.abc import Iterator
from pathlib import Path

import pytest
from astroid import extract_node, nodes

from pylint.pyreverse.diadefslib import (
    ClassDiadefGenerator,
    DefaultDiadefGenerator,
    DiaDefGenerator,
    DiadefsHandler,
)
from pylint.pyreverse.diagrams import DiagramEntity, Relationship
from pylint.pyreverse.inspector import Linker, Project
from pylint.testutils.pyreverse import PyreverseConfig
from pylint.testutils.utils import _test_cwd
from pylint.typing import GetProjectCallable

HERE = Path(__file__)
TESTS = HERE.parent.parent


def _process_classes(classes: list[DiagramEntity]) -> list[tuple[bool, str]]:
    """Extract class names of a list."""
    return sorted((isinstance(c.node, nodes.ClassDef), c.title) for c in classes)


def _process_relations(
    relations: dict[str, list[Relationship]]
) -> list[tuple[str, str, str]]:
    """Extract relation indices from a relation list."""
    result = []
    for rel_type, rels in relations.items():
        for rel in rels:
            result.append((rel_type, rel.from_object.title, rel.to_object.title))
    result.sort()
    return result


@pytest.fixture
def HANDLER(default_config: PyreverseConfig) -> DiadefsHandler:
    return DiadefsHandler(default_config)


@pytest.fixture(scope="module")
def PROJECT(get_project: GetProjectCallable) -> Iterator[Project]:
    with _test_cwd(TESTS):
        yield get_project("data")


def test_option_values(
    default_config: PyreverseConfig, HANDLER: DiadefsHandler, PROJECT: Project
) -> None:
    """Test for ancestor, associated and module options."""
    df_h = DiaDefGenerator(Linker(PROJECT), HANDLER)
    cl_config = default_config
    cl_config.classes = ["Specialization"]
    cl_h = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(cl_config))
    assert df_h._get_levels() == (0, 0)
    assert not df_h.module_names
    assert cl_h._get_levels() == (-1, -1)
    assert cl_h.module_names
    for hndl in (df_h, cl_h):
        hndl.config.all_ancestors = True
        hndl.config.all_associated = True
        hndl.config.module_names = True
        hndl._set_default_options()
        assert hndl._get_levels() == (-1, -1)
        assert hndl.module_names
    handler = DiadefsHandler(default_config)
    df_h = DiaDefGenerator(Linker(PROJECT), handler)
    cl_config = default_config
    cl_config.classes = ["Specialization"]
    cl_h = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(cl_config))
    for hndl in (df_h, cl_h):
        hndl.config.show_ancestors = 2
        hndl.config.show_associated = 1
        hndl.config.module_names = False
        hndl._set_default_options()
        assert hndl._get_levels() == (2, 1)
        assert not hndl.module_names


def test_default_values() -> None:
    """Test default values for package or class diagrams."""
    # TODO : should test difference between default values for package or class diagrams


class TestShowOptions:
    def test_show_stdlib(self) -> None:
        example = extract_node(
            '''
            import collections

            class CustomDict(collections.OrderedDict):
                """docstring"""
            '''
        )

        config = PyreverseConfig()
        dd_gen = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(config))

        # Default behavior
        assert not list(dd_gen.get_ancestors(example, 1))

        # Show standard library enabled
        config.show_stdlib = True
        ancestors = list(dd_gen.get_ancestors(example, 1))
        assert len(ancestors) == 1
        assert ancestors[0].name == "OrderedDict"

    def test_show_builtin(self) -> None:
        example = extract_node(
            '''
            class CustomError(Exception):
                """docstring"""
            '''
        )

        config = PyreverseConfig()
        dd_gen = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(config))

        # Default behavior
        assert not list(dd_gen.get_ancestors(example, 1))

        # Show builtin enabled
        config.show_builtin = True
        ancestors = list(dd_gen.get_ancestors(example, 1))
        assert len(ancestors) == 1
        assert ancestors[0].name == "Exception"


class TestDefaultDiadefGenerator:
    _should_rels = [
        ("aggregation", "DoNothing2", "Specialization"),
        ("association", "DoNothing", "Ancestor"),
        ("association", "DoNothing", "Specialization"),
        ("specialization", "Specialization", "Ancestor"),
    ]

    def test_extract_relations(self, HANDLER: DiadefsHandler, PROJECT: Project) -> None:
        """Test extract_relations between classes."""
        cd = DefaultDiadefGenerator(Linker(PROJECT), HANDLER).visit(PROJECT)[1]
        cd.extract_relationships()
        relations = _process_relations(cd.relationships)
        assert relations == self._should_rels

    def test_functional_relation_extraction(
        self, default_config: PyreverseConfig, get_project: GetProjectCallable
    ) -> None:
        """Functional test of relations extraction;
        different classes possibly in different modules
        """
        # XXX should be catching pyreverse environment problem but doesn't
        # pyreverse doesn't extract the relations but this test ok
        project = get_project("data")
        handler = DiadefsHandler(default_config)
        diadefs = handler.get_diadefs(project, Linker(project, tag=True))
        cd = diadefs[1]
        relations = _process_relations(cd.relationships)
        assert relations == self._should_rels


def test_known_values1(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    dd = DefaultDiadefGenerator(Linker(PROJECT), HANDLER).visit(PROJECT)
    assert len(dd) == 2
    keys = [d.TYPE for d in dd]
    assert keys == ["package", "class"]
    pd = dd[0]
    assert pd.title == "packages No Name"
    modules = sorted((isinstance(m.node, nodes.Module), m.title) for m in pd.objects)
    assert modules == [
        (True, "data"),
        (True, "data.clientmodule_test"),
        (True, "data.nullable_pattern"),
        (True, "data.property_pattern"),
        (True, "data.suppliermodule_test"),
    ]
    cd = dd[1]
    assert cd.title == "classes No Name"
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "Ancestor"),
        (True, "CustomException"),
        (True, "DoNothing"),
        (True, "DoNothing2"),
        (True, "DoSomething"),
        (True, "Interface"),
        (True, "NullablePatterns"),
        (True, "PropertyPatterns"),
        (True, "Specialization"),
    ]


def test_known_values2(
    HANDLER: DiadefsHandler, get_project: GetProjectCallable
) -> None:
    project = get_project("data.clientmodule_test")
    dd = DefaultDiadefGenerator(Linker(project), HANDLER).visit(project)
    assert len(dd) == 1
    keys = [d.TYPE for d in dd]
    assert keys == ["class"]
    cd = dd[0]
    assert cd.title == "classes No Name"
    classes = _process_classes(cd.objects)
    assert classes == [(True, "Ancestor"), (True, "Specialization")]


def test_known_values3(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    HANDLER.config.classes = ["Specialization"]
    cdg = ClassDiadefGenerator(Linker(PROJECT), HANDLER)
    special = "data.clientmodule_test.Specialization"
    cd = cdg.class_diagram(PROJECT, special)
    assert cd.title == special
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "data.clientmodule_test.Ancestor"),
        (True, special),
        (True, "data.suppliermodule_test.DoNothing"),
        (True, "data.suppliermodule_test.DoNothing2"),
    ]


def test_known_values4(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    HANDLER.config.classes = ["Specialization"]
    HANDLER.config.module_names = False
    cd = ClassDiadefGenerator(Linker(PROJECT), HANDLER).class_diagram(
        PROJECT, "data.clientmodule_test.Specialization"
    )
    assert cd.title == "data.clientmodule_test.Specialization"
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "Ancestor"),
        (True, "DoNothing"),
        (True, "DoNothing2"),
        (True, "Specialization"),
    ]


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Requires dataclasses")
def test_regression_dataclasses_inference(
    HANDLER: DiadefsHandler, get_project: GetProjectCallable
) -> None:
    project_path = Path("regrtest_data") / "dataclasses_pyreverse"
    path = get_project(str(project_path))

    cdg = ClassDiadefGenerator(Linker(path), HANDLER)
    special = "regrtest_data.dataclasses_pyreverse.InventoryItem"
    cd = cdg.class_diagram(path, special)
    assert cd.title == special



"""Tests for the useless-with-lock message"""
# pylint: disable=missing-docstring
import threading
from threading import Lock, RLock, Condition, Semaphore, BoundedSemaphore


with threading.Lock():  # [useless-with-lock]
    ...

with Lock():  # [useless-with-lock]
    ...

with threading.Lock() as this_shouldnt_matter:  # [useless-with-lock]
    ...

with threading.RLock():  # [useless-with-lock]
    ...

with RLock():  # [useless-with-lock]
    ...

with threading.Condition():  # [useless-with-lock]
    ...

with Condition():  # [useless-with-lock]
    ...

with threading.Semaphore():  # [useless-with-lock]
    ...

with Semaphore():  # [useless-with-lock]
    ...

with threading.BoundedSemaphore():  # [useless-with-lock]
    ...

with BoundedSemaphore():  # [useless-with-lock]
    ...

lock = threading.Lock()
with lock:  # this is ok
    ...

rlock = threading.RLock()
with rlock:  # this is ok
    ...

cond = threading.Condition()
with cond:  # this is ok
    ...

sem = threading.Semaphore()
with sem:  # this is ok
    ...

b_sem = threading.BoundedSemaphore()
with b_sem:  # this is ok
    ...



"""https://github.com/pylint-dev/pylint/issues/5371"""
from enum import Enum


class MyEnum(Enum):
    """
    Enum._generate_next_value_() in the stdlib currently lacks a
    @staticmethod decorator.
    """

    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list):
        return 42



"""Tests for logging-too-many-args"""
import logging

logging.error("constant string", 1, 2)  # [logging-too-many-args]
logging.error("{}", 1, 2)  # [logging-too-many-args]
logging.error("{0}", 1, 2)  # [logging-too-many-args]
logging.error("{}, {named}", 1, 2, {"named": 1})  # [logging-too-many-args]
logging.error("{0}, {named}", 1, 2, {"named": 1})  # [logging-too-many-args]



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt
import os
from pathlib import Path

import pytest

from pylint.pyreverse.main import Run
from pylint.testutils.pyreverse import (
    FunctionalPyreverseTestfile,
    get_functional_test_files,
)

FUNCTIONAL_DIR = Path(__file__).parent / "functional"
CLASS_DIAGRAMS_DIR = FUNCTIONAL_DIR / "class_diagrams"
CLASS_DIAGRAM_TESTS = get_functional_test_files(CLASS_DIAGRAMS_DIR)
CLASS_DIAGRAM_TEST_IDS = [testfile.source.stem for testfile in CLASS_DIAGRAM_TESTS]


@pytest.mark.parametrize(
    "testfile",
    CLASS_DIAGRAM_TESTS,
    ids=CLASS_DIAGRAM_TEST_IDS,
)
def test_class_diagrams(testfile: FunctionalPyreverseTestfile, tmp_path: Path) -> None:
    input_file = testfile.source
    input_path = os.path.dirname(input_file)
    if testfile.options["source_roots"]:
        source_roots = ",".join(
            [
                os.path.realpath(
                    os.path.expanduser(os.path.join(input_path, source_root))
                )
                for source_root in testfile.options["source_roots"]
            ]
        )
    else:
        source_roots = ""
    for output_format in testfile.options["output_formats"]:
        with pytest.raises(SystemExit) as sys_exit:
            args = ["-o", f"{output_format}", "-d", str(tmp_path)]
            if source_roots:
                args += ["--source-roots", source_roots]
            args.extend(testfile.options["command_line_args"])
            args += [str(input_file)]
            Run(args)
        assert sys_exit.value.code == 0
        assert testfile.source.with_suffix(f".{output_format}").read_text(
            encoding="utf8"
        ) == (tmp_path / f"classes.{output_format}").read_text(encoding="utf8")



"""Test for multi naming style."""


def UPPER():  # [invalid-name]
    """UPPER func that will generate invalid-name error."""


def lower():
    """lower func that satisfies the prevalent group `lower`."""


def anotherlower():
    """Another lower func that satisfies the prevalent group `lower`."""



"""If the else block continues, it is generally safe to rely on assignments in the except,
inside the same for loop only."""


def safe():
    """Name used safely inside the loop."""
    while True:
        try:
            pass
        except ValueError:
            error = True
        else:
            continue

        print(error)


def halfway_safe():
    """Name used safely inside the loop, unsafely outside it."""
    for _temp in range(0, 1):
        try:
            pass
        except ValueError:
            error = True
        else:
            continue

        print(error)
    print(error)  # https://github.com/pylint-dev/pylint/issues/9379


def unsafe():
    """Name used unsafely outside the loop."""
    for _temp in range(0, 1):
        try:
            pass
        except ValueError:
            error = True
        else:
            continue

    print(error)  # [used-before-assignment]



"""Check invalid value returned by __bytes__ """

# pylint: disable=too-few-public-methods,missing-docstring,import-error,unnecessary-lambda-assignment
import six

from missing import Missing


class FirstGoodBytes:
    """__bytes__ returns <type 'bytes'>"""

    def __bytes__(self):
        return b"some bytes"


class SecondGoodBytes:
    """__bytes__ returns <type 'bytes'>"""

    def __bytes__(self):
        return bytes("123", "ascii")


class BytesMetaclass(type):
    def __bytes__(cls):
        return b"some bytes"


@six.add_metaclass(BytesMetaclass)
class ThirdGoodBytes:
    """Bytes through the metaclass."""


class FirstBadBytes:
    """ __bytes__ returns bytes """

    def __bytes__(self):  # [invalid-bytes-returned]
        return "123"


class SecondBadBytes:
    """ __bytes__ returns int """

    def __bytes__(self):  # [invalid-bytes-returned]
        return 1


class ThirdBadBytes:
    """ __bytes__ returns node which does not have 'value' in AST """

    def __bytes__(self):  # [invalid-bytes-returned]
        return lambda: b"some bytes"


class AmbiguousBytes:
    """ Uninferable return value """

    __bytes__ = lambda self: Missing


class AnotherAmbiguousBytes:
    """Potential uninferable return value"""

    def __bytes__(self):
        return bytes(Missing)



# pylint: disable=missing-docstring,invalid-name,too-few-public-methods

class Point2D:
    __match_args__ = ("x", "y")

    def __init__(self, x, y):
        self.x = x
        self.y = y


# Don't emit false-positive 'unused-variable' and 'undefined-variable'
var = 42
match var:
    case [*rest1]:
        print(rest1)
    case {**rest2}:
        print(rest2)
    case Point2D(0, a):
        print(a)
    case Point2D(x=0, y=b) as new_point:
        print(b)
        print(new_point)
    case new_var:
        print(new_var)


# Test inference of variables assigned in patterns doesn't crash pylint
var = 42
match var:
    case (1, *rest3):
        if rest3 != 2:
            pass
    case {1: _, **rest4}:
        if rest4 != 2:
            pass
    case c:
        if c != 2:
            pass



"""
Example of trojan unicode, see https://trojansource.codes/
This example was taken from PEP672
"""
# pylint: disable=invalid-name

# +1: [bidirectional-unicode]
example = "x‏" * 100  #    "‏x" is assigned



# pylint: disable=missing-docstring,unused-argument,pointless-statement
# pylint: disable=too-few-public-methods,no-name-in-module
from __future__ import annotations as __annotations__


class MyClass:
    @classmethod
    def from_string(cls, source) -> MyClass:
        ...

    def validate_b(self, obj: OtherClass) -> bool:
        ...


class OtherClass:
    ...


class Example:
    obj: Other


class Other:
    ...


class ExampleSelf:
    next: ExampleSelf



# pylint: disable=too-few-public-methods, invalid-name
"""Test typing with a protected member"""
from __future__ import annotations


class MyClass:
    """Class with protected members."""

    class _Inner_Class:
        """Inner class with protected members."""

        def __init__(self) -> None:
            self.data = 1

        def return_data(self) -> int:
            """Return data"""
            return self.data

    def return_private_class(self) -> MyClass._Inner_Class:
        """Doing nothing."""
        return self._Inner_Class()


def access_protected_class(data: MyClass._Inner_Class) -> int:
    """Function that always receives a protected class."""
    return data.return_data() + 1


def pass_protected_class() -> None:
    """Function that passes a protected class to another function."""
    data_value = access_protected_class(MyClass().return_private_class())
    print(data_value)



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import contextlib
import io
import os
import sys
import warnings
from collections.abc import Iterator

import pytest
from pytest import CaptureFixture

from pylint.testutils._run import _Run as Run


def is_module(filename: str) -> bool:
    return filename.endswith(".py")


def is_package(filename: str, location: str) -> bool:
    return os.path.exists(os.path.join(location, filename, "__init__.py"))


@contextlib.contextmanager
def _patch_stdout(out: io.StringIO) -> Iterator[None]:
    sys.stdout = out
    try:
        yield
    finally:
        sys.stdout = sys.__stdout__


LIB_DIRS = [os.path.dirname(os.__file__)]
MODULES_TO_CHECK = [
    (location, module)
    for location in LIB_DIRS
    for module in os.listdir(location)
    if is_module(module) or is_package(module, location)
]
MODULES_NAMES = [m[1] for m in MODULES_TO_CHECK]


@pytest.mark.primer_stdlib
@pytest.mark.parametrize(
    ("test_module_location", "test_module_name"), MODULES_TO_CHECK, ids=MODULES_NAMES
)
def test_primer_stdlib_no_crash(
    test_module_location: str, test_module_name: str, capsys: CaptureFixture
) -> None:
    """Test that pylint does not produce any crashes or fatal errors on stdlib modules."""
    __tracebackhide__ = True  # pylint: disable=unused-variable
    os.chdir(test_module_location)
    with _patch_stdout(io.StringIO()):
        try:
            # We want to test all the code we can
            enables = ["--enable-all-extensions", "--enable=all"]
            # Duplicate code takes too long and is relatively safe
            # We don't want to lint the test directory which are repetitive
            disables = ["--disable=duplicate-code", "--ignore=test"]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                Run([test_module_name, *enables, *disables])
        except SystemExit as ex:
            out, err = capsys.readouterr()
            assert not err, err
            assert not out
            msg = f"Encountered {{}} during primer stlib test for {test_module_name}"
            assert ex.code != 32, msg.format("a crash")
            assert ex.code % 2 == 0, msg.format("a message of category 'fatal'")  # type: ignore[operator]



"""check builtin data descriptors such as mode and name attributes on a file are correctly handled

bug notified by Pierre Rouleau on 2005-04-24
"""


class File(file):  # pylint: disable=undefined-variable
    """ Testing new-style class inheritance from file"""
    def __init__(self, name, mode="r", buffering=-1, verbose=False):
        """Constructor"""
        self.was_modified = False
        self.verbose = verbose
        super(File, self).__init__(name, mode, buffering)  # [super-with-arguments]
        if self.verbose:
            print(f"File {self.name} is opened.  The mode is: {self.mode}")

    def write(self, a_string):
        """ Write a string to the file."""
        super(File, self).write(a_string)  # [super-with-arguments]
        self.was_modified = True

    def writelines(self, sequence):
        """ Write a sequence of strings to the file. """
        super(File, self).writelines(sequence)  # [super-with-arguments]
        self.was_modified = True

    def close(self):
        """Close the file."""
        if self.verbose:
            print(f"Closing file {self.name}")
        super(File, self).close()  # [super-with-arguments]
        self.was_modified = False



"""Emit a message for breaking out of a while True loop immediately."""
# pylint: disable=missing-function-docstring,missing-class-docstring,unrecognized-inline-option,invalid-name,literal-comparison, undefined-variable, too-many-public-methods, no-else-break

class Issue8015:
    def bad(self):
        k = 1
        while 1:  # [consider-refactoring-into-while-condition]
            if k == 10:
                break
            k += 1

    def another_bad(self):
        current_scope = None
        while 2:  # [consider-refactoring-into-while-condition]
            if current_scope is None:
                break
            current_scope = True

    def good(self):
        k = 1
        while True:
            k += 1
            if k == 10:
                break

    def another_good(self):
        k = 1
        while k < 10:
            k += 1

    def test_error_message_multiple_break(self, k: int) -> None:
        while True:  # [consider-refactoring-into-while-condition]
            if k <= 1:
                break
            if k > 10:
                break
            k -= 1

    def test_error_message(self):
        a_list = [1,2,3,4,5]
        # Should recommend `while a_list`
        while True:   # [consider-refactoring-into-while-condition]
            if not a_list:
                break
            a_list.pop()

    def test_error_message_2(self):
        a_list = []
        # Should recommend `while not a_list`
        while True:   # [consider-refactoring-into-while-condition]
            if a_list:
                break
            a_list.append(1)

    def test_error_message_3(self):
        a_var = "defined"
        # Should recommend `while not a_var`
        while True:   # [consider-refactoring-into-while-condition]
            if a_var is not None:
                break
            a_var = None

    def test_error_message_4(self):
        a_list = []
        # Should recommend `while a_list is []`
        while True:   # [consider-refactoring-into-while-condition]
            if a_list is not []:
                break
            a_list.append(1)

    def test_error_message_5(self):
        a_dict = {}
        a_var = a_dict.get("undefined_key")
        # Should recommend `while a_var`
        while True:   # [consider-refactoring-into-while-condition]
            if a_var is None:
                break
            a_var = "defined"

    def test_error_message_6(self):
        a_list = []
        # Should recommend `while a_list is not []`
        while True:   # [consider-refactoring-into-while-condition]
            if a_list is []:
                break
            a_list.append(1)


    def test_error_message_7(self):
        # while not a and b is not correct
        # Expeccted message should be `while not (a and b)``
        a = True
        b = False
        while True:   # [consider-refactoring-into-while-condition]
            if a and b:
                break
            a = 1
            b = 1

    def test_error_message_8(self):
        # while not a and not b is not correct
        # Expeccted message should be `while not (a and not b)``
        a = True
        b = False
        while True:   # [consider-refactoring-into-while-condition]
            if a and not b:
                break
            a = 1
            b = 1

    def test_error_message_9(self):
        k = 1
        while True:  # [consider-refactoring-into-while-condition]
            if k != 1:
                break
            k += 1

    def test_error_message_10(self):
        a = [1,2,3,4,5]
        while True:  # [consider-refactoring-into-while-condition]
            if 5 not in a:
                break
            a.pop()

    def test_error_message_11(self):
        a = []
        k = 1
        while True:  # [consider-refactoring-into-while-condition]
            if 5 in a:
                break
            a.append(k)
            k += 1

    def test_error_message_12(self):
        k = 1
        while True:  # [consider-refactoring-into-while-condition]
            if k > 10:
                break
            k += 1

    def test_error_message_13(self):
        k = 1
        while True:  # [consider-refactoring-into-while-condition]
            if k >= 10:
                break
            k += 1

    def test_error_message_14(self):
        k = 10
        while True:  # [consider-refactoring-into-while-condition]
            if k < 1:
                break
            k -= 1

    def test_error_message_15(self):
        k = 1
        while True:  # [consider-refactoring-into-while-condition]
            if k <= 1:
                break
            k -= 1

    def test_error_message_16(self):
        # Silly example but needed for coverage
        k = None
        while True:  # [consider-refactoring-into-while-condition]
            if (lambda x: x) == k:
                break
            break
        while True:  # [consider-refactoring-into-while-condition]
            if k == (lambda x: x):
                break
            break

    def test_error_message_17(self):
        a = True
        b = False
        c = True
        d = False
        while True:  # [consider-refactoring-into-while-condition]
            if (a or b) == (c and d):
                break
            a = not a if random.randint(0,1) == 1 else a
            b = not b if random.randint(0,1) == 1 else b
            c = not c if random.randint(0,1) == 1 else c
            d = not d if random.randint(0,1) == 1 else d

        while True:  # [consider-refactoring-into-while-condition]
            if (not a) == (not d):
                break
            a = not a if random.randint(0,1) == 1 else a
            d = not d if random.randint(0,1) == 1 else d

    def test_error_message_18(self):
        x = 0
        while True:  # [consider-refactoring-into-while-condition]
            if x ** 2:
                break
            x += 1

    def test_multi_break_condition_1(self):
        x = 0
        # This should chain conditions into
        # While (x == 0) and (x >= 0) and (x != 0):
        while True:  # [consider-refactoring-into-while-condition]
            if x != 0:
                break
            elif x > 0:
                x -= 1
            elif x < 0:
                break
            elif x == 0:
                break
            x -= 10

    def test_multi_break_condition_2(self):
        x = 0
        # This should chain both conditions
        while True:  # [consider-refactoring-into-while-condition]
            if x != 0:
                break
            if x == 0:
                break
            x -= 10

    def test_multi_break_condition_3(self):
        x = 0
        # This should chain all conditions
        while True:  # [consider-refactoring-into-while-condition]
            if x != 0:
                break
            elif x < 0:
                break
            elif x == 0:
                break
            if x != 100:
                break
            if x == 1000:
                break
            x -= 10

    def test_multi_break_condition_4(self):
        x = 0
        # This should chain all conditions except last 2.
        # The else clause taints the first if-elif-else block by introducing mutation
        while True:  # [consider-refactoring-into-while-condition]
            if x != 0:
                break
            elif x < 0:
                break
            elif x == 0:
                break
            else:
                x += 1
            if x != 100:
                break
            if x == 1000:
                break
            x -= 10

    def falsy_1(self):
        x = 0
        while []:
            if x > 10:
                break
            x += 1

    def falsy_2(self):
        x = 0
        while ():
            if x > 10:
                break
            x += 1

    def falsy_3(self):
        x = 0
        while {}:
            if x > 10:
                break
            x += 1

    def falsy_4(self):
        x = 0
        while set():
            if x > 10:
                break
            x += 1

    def falsy_5(self):
        x = 0
        while "":
            if x > 10:
                break
            x += 1

    def falsy_6(self):
        x = 0
        while range(0):
            if x > 10:
                break
            x += 1

    def falsy_7(self):
        x = 0
        while 0:
            if x > 10:
                break
            x += 1

    def falsy_8(self):
        x = 0
        while 0.0:
            if x > 10:
                break
            x += 1

    def falsy_9(self):
        x = 0
        while 0j:
            if x > 10:
                break
            x += 1

    def falsy_10(self):
        x = 0
        while None:
            if x > 10:
                break
            x += 1

    def falsy_11(self):
        x = 0
        while False:
            if x > 10:
                break
            x += 1



""" Tests for invalid-name checker in the context of enums. """
# pylint: disable=too-few-public-methods


from dataclasses import dataclass
from enum import Enum


class Color(Enum):
    """Represents colors as (red, green, blue) tuples."""

    YELLOW      = 250, 250,   0
    KHAKI       = 250, 250, 125
    MAGENTA     = 250,   0, 250
    VIOLET      = 250, 125, 250
    CYAN        =   0, 250, 250
    aquamarine  = 125, 250, 250  # [invalid-name]

    red: int
    green: int
    blue: int

    def __init__(self, red: int, green: int, blue: int) -> None:
        self.red   = red
        self.green = green
        self.blue  = blue

    @property
    def as_hex(self) -> str:
        """Get hex 'abcdef' representation for a color."""
        return f'{self.red:0{2}x}{self.green:0{2}x}{self.blue:0{2}x}'


@dataclass
class Something(str, Enum):
    """ A false positive for ``invalid-name``
    which should be fixed by https://github.com/pylint-dev/astroid/issues/2317
    """
    ASD: str = 1  # [invalid-name]



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Unittest for the NameChecker."""

from __future__ import annotations

import unittest

from pylint.checkers import base


class TestNamePresets(unittest.TestCase):
    SNAKE_CASE_NAMES = {"tést_snake_case", "test_snake_case11", "test_https_200"}
    CAMEL_CASE_NAMES = {"téstCamelCase", "testCamelCase11", "testHTTP200"}
    UPPER_CASE_NAMES = {"TÉST_UPPER_CASE", "TEST_UPPER_CASE11", "TEST_HTTP_200"}
    PASCAL_CASE_NAMES = {"TéstPascalCase", "TestPascalCase11", "TestHTTP200"}
    ALL_NAMES = (
        SNAKE_CASE_NAMES | CAMEL_CASE_NAMES | UPPER_CASE_NAMES | PASCAL_CASE_NAMES
    )

    def _test_name_is_correct_for_all_name_types(
        self, naming_style: type[base.NamingStyle], name: str
    ) -> None:
        for name_type in base.KNOWN_NAME_TYPES_WITH_STYLE:
            self._test_is_correct(naming_style, name, name_type)

    def _test_name_is_incorrect_for_all_name_types(
        self, naming_style: type[base.NamingStyle], name: str
    ) -> None:
        for name_type in base.KNOWN_NAME_TYPES_WITH_STYLE:
            self._test_is_incorrect(naming_style, name, name_type)

    def _test_should_always_pass(self, naming_style: type[base.NamingStyle]) -> None:
        always_pass_data = [
            ("__add__", "method"),
            ("__set_name__", "method"),
            ("__version__", "const"),
            ("__author__", "const"),
        ]
        for name, name_type in always_pass_data:
            self._test_is_correct(naming_style, name, name_type)

    @staticmethod
    def _test_is_correct(
        naming_style: type[base.NamingStyle], name: str, name_type: str
    ) -> None:
        rgx = naming_style.get_regex(name_type)
        fail = f"{name!r} does not match pattern {rgx!r} (style: {naming_style}, type: {name_type})"
        assert rgx.match(name), fail

    @staticmethod
    def _test_is_incorrect(
        naming_style: type[base.NamingStyle], name: str, name_type: str
    ) -> None:
        rgx = naming_style.get_regex(name_type)
        fail = f"{name!r} not match pattern {rgx!r} (style: {naming_style}, type: {name_type})"
        assert not rgx.match(name), fail

    def test_snake_case(self) -> None:
        naming_style = base.SnakeCaseStyle

        for name in self.SNAKE_CASE_NAMES:
            self._test_name_is_correct_for_all_name_types(naming_style, name)
        for name in self.ALL_NAMES - self.SNAKE_CASE_NAMES:
            self._test_name_is_incorrect_for_all_name_types(naming_style, name)

        self._test_should_always_pass(naming_style)

    def test_camel_case(self) -> None:
        naming_style = base.CamelCaseStyle

        for name in self.CAMEL_CASE_NAMES:
            self._test_name_is_correct_for_all_name_types(naming_style, name)
        for name in self.ALL_NAMES - self.CAMEL_CASE_NAMES:
            self._test_name_is_incorrect_for_all_name_types(naming_style, name)

        self._test_should_always_pass(naming_style)

    def test_upper_case(self) -> None:
        naming_style = base.UpperCaseStyle

        for name in self.UPPER_CASE_NAMES:
            self._test_name_is_correct_for_all_name_types(naming_style, name)
        for name in self.ALL_NAMES - self.UPPER_CASE_NAMES:
            self._test_name_is_incorrect_for_all_name_types(naming_style, name)
        self._test_name_is_incorrect_for_all_name_types(naming_style, "UPPERcase")

        self._test_should_always_pass(naming_style)

    def test_pascal_case(self) -> None:
        naming_style = base.PascalCaseStyle

        for name in self.PASCAL_CASE_NAMES:
            self._test_name_is_correct_for_all_name_types(naming_style, name)
        for name in self.ALL_NAMES - self.PASCAL_CASE_NAMES:
            self._test_name_is_incorrect_for_all_name_types(naming_style, name)

        self._test_should_always_pass(naming_style)



"""Tests for shallow-copy-environ"""
# pylint: disable=wrong-import-position, reimported, ungrouped-imports, import-error, wrong-import-order

import copy
import os

copy.copy(os.environ)  # [shallow-copy-environ]

# Copying a dictionary is okay
test_dict = {}
copy.copy(test_dict)

# Test with renaming the functions
from copy import copy as test_cp
import os as o

test_cp(o.environ)  # [shallow-copy-environ]

# Regression test for non inferable objects
import copy
from missing_library import MissingObject

copy.copy(MissingObject)

# Deepcopy is okay
import copy
import os

copy.deepcopy(os.environ)

# Edge cases
copy.copy()  # [no-value-for-parameter]
copy.copy(x=test_dict)
copy.copy(x=os.environ)  # [shallow-copy-environ]
copy.copy(**{"x": os.environ})  # [shallow-copy-environ]
copy.copy(**{"y": os.environ})  # [unexpected-keyword-arg]
copy.copy(y=os.environ)  # [no-value-for-parameter, unexpected-keyword-arg]



"""Test PEP 585 without postponed evaluation. Everything should fail.

This check requires Python 3.7 or Python 3.8!
Testing with 3.8 only, to support TypedDict.
"""

# pylint: disable=missing-docstring,unused-argument,unused-import,too-few-public-methods
# pylint: disable=invalid-name,inherit-non-class,unsupported-binary-operation
# pylint: disable=unused-variable,line-too-long,unnecessary-direct-lambda-call

# Disabled because of a bug with pypy 3.8 see
# https://github.com/pylint-dev/pylint/pull/7918#issuecomment-1352737369
# pylint: disable=multiple-statements

import collections
import dataclasses
import typing
from dataclasses import dataclass
from typing import Any, Dict, NamedTuple, TypedDict, Union


AliasInvalid = list[int]  # [unsubscriptable-object]

class CustomIntList(typing.List[int]):
    pass

class CustomIntListError(list[int]):  # [unsubscriptable-object]
    pass

cast_variable = [1, 2, 3]
cast_variable = typing.cast(list[int], cast_variable)  # [unsubscriptable-object]

T = typing.TypeVar("T", list[int], str)  # [unsubscriptable-object]

(lambda x: 2)(list[int])  # [unsubscriptable-object]


# Check typing.NamedTuple
CustomNamedTuple = typing.NamedTuple(
    "CustomNamedTuple", [("my_var", list[int])])  # [unsubscriptable-object]

class CustomNamedTuple2(NamedTuple):
    my_var: list[int]  # [unsubscriptable-object]

class CustomNamedTuple3(typing.NamedTuple):
    my_var: list[int]  # [unsubscriptable-object]


# Check typing.TypedDict
CustomTypedDict = TypedDict("CustomTypedDict", my_var=list[int])  # [unsubscriptable-object]

CustomTypedDict2 = TypedDict("CustomTypedDict2", {"my_var": list[int]})  # [unsubscriptable-object]

class CustomTypedDict3(TypedDict):
    my_var: list[int]  # [unsubscriptable-object]

class CustomTypedDict4(typing.TypedDict):
    my_var: list[int]  # [unsubscriptable-object]


# Check dataclasses
def my_decorator(*args, **kwargs):
    def wraps(*args, **kwargs):
        pass
    return wraps

@dataclass
class CustomDataClass:
    my_var: list[int]  # [unsubscriptable-object]

@dataclasses.dataclass
class CustomDataClass2:
    my_var: list[int]  # [unsubscriptable-object]

@dataclass()
class CustomDataClass3:
    my_var: list[int]  # [unsubscriptable-object]

@my_decorator
@dataclasses.dataclass
class CustomDataClass4:
    my_var: list[int]  # [unsubscriptable-object]


var1: set[int]  # [unsubscriptable-object]
var2: collections.OrderedDict[str, int]  # [unsubscriptable-object]
var3: dict[str, list[int]]  # [unsubscriptable-object,unsubscriptable-object]
var4: Dict[str, list[int]]  # [unsubscriptable-object]
var5: dict[tuple[int, int], str]  # [unsubscriptable-object,unsubscriptable-object]
var6: Dict[tuple[int, int], str]  # [unsubscriptable-object]
var7: list[list[int]]  # [unsubscriptable-object,unsubscriptable-object]
var8: tuple[list[int]]  # [unsubscriptable-object,unsubscriptable-object]
var9: int | list[str | int]   # [unsubscriptable-object]
var10: Union[list[str], None]   # [unsubscriptable-object]
var11: Union[Union[list[int], int]]   # [unsubscriptable-object]

def func(arg: list[int]):  # [unsubscriptable-object]
    pass

def func2() -> list[int]:  # [unsubscriptable-object]
    pass

Alias2 = Union[list[str], None]  # [unsubscriptable-object]
Alias3 = Union[Union[list[int], int]]  # [unsubscriptable-object]
Alias5 = Dict[str, list[int]]  # [unsubscriptable-object]
Alias6 = int | list[int]  # [unsubscriptable-object]
Alias7 = list[list[int]]  # [unsubscriptable-object,unsubscriptable-object]


def func3():
    AliasInvalid2 = list[int]  # [unsubscriptable-object]
    cast_variable2 = [1, 2, 3]
    cast_variable2 = typing.cast(list[int], cast_variable2)  # [unsubscriptable-object]
    var12: list[int]  # [unsubscriptable-object]

def func4(var=list[int]):  # [unsubscriptable-object]
    pass

def func5(arg1: list[int], arg2=set[int]):  # [unsubscriptable-object,unsubscriptable-object]
    pass

def func6(arg1: list[int], /, *args: tuple[str], arg2: set[int], **kwargs: dict[str, Any]):
    # -1:[unsubscriptable-object,unsubscriptable-object,unsubscriptable-object,unsubscriptable-object]
    pass



# pylint: disable=missing-module-docstring, invalid-name, missing-class-docstring, unused-variable


from dataclasses import dataclass
from typing import Generic, TypeVar


T_Inner = TypeVar("T_Inner", bound="Inner")


@dataclass
class Inner:
    inner_attribute: str


@dataclass
class Outer(Generic[T_Inner]):
    inner: T_Inner


x = Outer(inner=Inner(inner_attribute="magic xylophone"))

# Test `no-member` is not emitted here.
# https://github.com/pylint-dev/pylint/issues/9069
print(x.inner.inner_attribute)



"""Tests for undefined-variable related to classes"""
# pylint: disable=missing-function-docstring, missing-class-docstring, too-few-public-methods

# Test that list comprehensions in base classes are scoped correctly
# Regression reported in https://github.com/pylint-dev/pylint/issues/3434

import collections

l = ["a", "b", "c"]


class Foo(collections.namedtuple("Foo", [x + "_foo" for x in l])):
    pass


# Test that class attributes are in scope for return type annotations.
# Regression reported in https://github.com/pylint-dev/pylint/issues/1976
class MyObject:
    class MyType:
        pass

    def my_method(self) -> MyType:
        pass



# pylint: disable=missing-docstring,consider-using-with,redefined-outer-name

import pytest


@pytest.fixture
def qm_file():
    qm_file = open("src/test/resources/example_qm_file.csv", encoding="utf-8").read()
    return qm_file



# pylint: disable=pointless-statement,missing-docstring
# https://github.com/pylint-dev/pylint/issues/4358

from enum import Enum

class Foo(Enum):
    BAR = 1

Foo.__members__['BAR']  # <-- error: Value 'Foo.__members__' is unsubscriptable



"""
InitVars should not count as instance attributes (see issue #3754)
Default max_instance_attributes is 7
"""

# pylint: disable=missing-docstring, too-few-public-methods

# Disabled because of a bug with pypy 3.8 see
# https://github.com/pylint-dev/pylint/pull/7918#issuecomment-1352737369
# pylint: disable=multiple-statements

from dataclasses import dataclass, InitVar


@dataclass
class Hello:
    a_1: int
    a_2: int
    a_3: int
    a_4: int
    a_5: int
    a_6: int
    a_7: int
    a_8: InitVar[int]

    def __post_init__(self, a_8):
        self.a_1 += a_8



"""Tests for used-before-assignment with assignments in except handlers after
try blocks with return statements.
See: https://github.com/pylint-dev/pylint/issues/5500.
"""
# pylint: disable=inconsistent-return-statements,broad-exception-raised


def function():
    """Assume except blocks execute if the try block returns."""
    try:
        success_message = "success message"
        return success_message
    except ValueError:
        failure_message = "failure message"
    finally:
        print(failure_message)  # [used-before-assignment]

    return failure_message


def func_ok(var):
    """'msg' is defined in all ExceptHandlers."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        msg = "Attribute not defined"
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)


def func_ok2(var):
    """'msg' is defined in all ExceptHandlers that don't raise an Exception."""
    try:
        return 1 / var.some_other_func()
    except AttributeError as ex:
        raise Exception from ex
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)


def func_ok3(var):
    """'msg' is defined in all ExceptHandlers that don't return."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        return
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)


def func_ok4(var):
    """Define "msg" with a chained assignment."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        msg2 = msg = "Division by 0"
        print(msg2)
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)


def func_ok5(var):
    """Define 'msg' via unpacked iterable."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        msg, msg2 = ["Division by 0", "Division by 0"]
        print(msg2)
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)


def func_ok6(var):
    """Define 'msg' in one handler nested under if block."""
    err_message = "Division by 0"
    try:
        return 1 / var.some_other_func()
    except ZeroDivisionError:
        if err_message:
            msg = "Division by 0"
        else:
            msg = None
    print(msg)


def func_ok7(var):
    """Define 'msg' in one handler nested under with statement."""
    try:
        return 1 / var.some_other_func()
    except ZeroDivisionError:
        with open(__file__, encoding='utf-8') as my_file:
            msg = "Division by 0"
            my_file.write(msg)
    print(msg)


def func_ok8(var):
    """Define 'msg' in one handler via type annotation."""
    try:
        return 1 / var.some_other_func()
    except ZeroDivisionError:
        # See func_invalid2() for mere annotation without value
        msg: str = "Division by 0"
    print(msg)


def func_invalid1(var):
    """'msg' is not defined in one handler."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        pass
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)  # [used-before-assignment]


def func_invalid2(var):
    """'msg' is not defined in one handler."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        msg: str
    except ZeroDivisionError:
        msg = "Division by 0"
    print(msg)  # [used-before-assignment]


def func_invalid3(var):
    """'msg' is not defined in one handler, but is defined in another
    nested under an if. Nesting under an if tests that the implementation
    does not assume direct parentage between `msg=` and `except`, and
    the prior except is necessary to raise the message.
    """
    err_message = False
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        pass
    except ZeroDivisionError:
        if err_message:
            msg = "Division by 0"
        else:
            msg = None
    print(msg)  # [used-before-assignment]


def func_invalid4(var):
    """Define 'msg' in one handler nested under with statement."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        pass
    except ZeroDivisionError:
        with open(__file__, encoding='utf-8') as my_file:
            msg = "Division by 0"
            my_file.write("****")
    print(msg)  # [used-before-assignment]


def func_invalid5(var):
    """Define 'msg' in one handler only via chained assignment."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        pass
    except ZeroDivisionError:
        msg2 = msg = "Division by 0"
        print(msg2)
    print(msg)  # [used-before-assignment]


def func_invalid6(var):
    """Define 'msg' in one handler only via unpacked iterable."""
    try:
        return 1 / var.some_other_func()
    except AttributeError:
        pass
    except ZeroDivisionError:
        msg, msg2 = ["Division by 0"] * 2
        print(msg2)
    print(msg)  # [used-before-assignment]



"""Tests for self-defined Enum members (https://github.com/pylint-dev/pylint/issues/6805)"""
# pylint: disable=missing-docstring
# pylint: disable=too-few-public-methods
from enum import IntEnum


class Foo(type):
    pass


class Parent:
    def __new__(cls, *_args, **_kwargs):
        return object.__new__(cls)


class NotEnumHasDynamicGetAttrMetaclass(metaclass=Foo):
    def __new__(cls):
        return Parent.__new__(cls)

    def __getattr__(self, item):
        return item

    def magic(self):
        return self.dynamic


NotEnumHasDynamicGetAttrMetaclass().magic()


class Day(IntEnum):
    MONDAY = (1, "Mon")
    TUESDAY = (2, "Tue")
    WEDNESDAY = (3, "Wed")
    THURSDAY = (4, "Thu")
    FRIDAY = (5, "Fri")
    SATURDAY = (6, "Sat")
    SUNDAY = (7, "Sun")

    def __new__(cls, value, _abbr=None):
        return int.__new__(cls, value)


print(Day.FRIDAY.foo)  # [no-member]



"""Tests for redefinitions of loop variables inside the loop body.

See: https://github.com/pylint-dev/pylint/issues/5608
"""
# pylint: disable=invalid-name

lines = ["1\t", "2\t"]
for line in lines:
    line = line.strip()  # [redefined-loop-name]

for i in range(8):
    for j in range(8):
        for i in range(j):  # [redefined-loop-name]
            j = i  # [redefined-loop-name]
            print(i, j)

for i in range(8):
    for j in range(8):
        for k in range(j):
            k = (i, j)  # [redefined-loop-name]
            print(i, j, k)

lines = [(1, "1\t"), (2, "2\t")]
for i, line in lines:
    line = line.strip()  # [redefined-loop-name]

line = "no warning"

for i in range(10):
    i += 1  # [redefined-loop-name]


def outer():
    """No redefined-loop-name for variables in nested scopes"""
    for i1 in range(5):
        def inner():
            """No warning, because i has a new scope"""
            for i1 in range(3):
                print(i1)
        print(i1)
        inner()


def outer2():
    """Similar, but with an assignment instead of homonymous loop variables"""
    for i1 in range(5):
        def inner():
            """No warning, because i has a new scope"""
            for _ in range(3):
                i1 = 0
                print(i1)
        print(i1)
        inner()



"""Test assignment of lambda expressions to a variable."""
# pylint: disable=unbalanced-tuple-unpacking, undefined-variable, line-too-long

a = lambda x: x  # [unnecessary-lambda-assignment]
a = (
    lambda x: x  # [unnecessary-lambda-assignment]
)
a, b = lambda x: x, lambda y: y  # [unnecessary-lambda-assignment,unnecessary-lambda-assignment]
a, b = (
    lambda x: x,  # [unnecessary-lambda-assignment]
    lambda y: y,  # [unnecessary-lambda-assignment]
)
a, b = 1, lambda y: y  # [unnecessary-lambda-assignment]
a, b = (
    1,
    lambda y: y  # [unnecessary-lambda-assignment]
)

# Interaction with W0632 (unbalanced-tuple-unpacking)
a, b = lambda x: x, lambda y: y, lambda z: z  # [unnecessary-lambda-assignment,unnecessary-lambda-assignment]
a, b = (
    lambda x: x,  # [unnecessary-lambda-assignment]
    lambda y: y,  # [unnecessary-lambda-assignment]
    lambda z: z,  # This isn't assigned so don't flag.
)
a, b, c = lambda x: x, lambda y: y  # [unnecessary-lambda-assignment,unnecessary-lambda-assignment]

# Only flag lambdas directly assigned to variables.
d["key"] = lambda x: x
squares = list(map(lambda x: x**2, range(10)))

DICT = {1: lambda x: x, 2: lambda x: x + 1}
for key, value in DICT.items():
    print(value(key))

# Flag lambda expression assignments via named expressions as well.
if (e := lambda: 2) and e():  # [unnecessary-lambda-assignment]
    pass



"""inline loop non ascii variable definition"""
import os


test = [
    f"{łol}                                                                       "
    for łol in os.listdir(".")  # [non-ascii-name]
]



"""
'collections.abc.Callable' is broken inside Optional and Union types for Python 3.9.0
https://bugs.python.org/issue42965

Use 'typing.Callable' instead.
"""
# pylint: disable=missing-docstring,unsubscriptable-object,invalid-name
import collections.abc
from collections.abc import Callable
from typing import TYPE_CHECKING, Optional, Union

Alias1 = Optional[Callable[[int], None]]  # [broken-collections-callable]
Alias2 = Union[Callable[[int], None], None]  # [broken-collections-callable]

Alias3 = Optional[Callable[..., None]]
Alias4 = Union[Callable[..., None], None]
Alias5 = list[Callable[..., None]]
Alias6 = Callable[[int], None]

if TYPE_CHECKING:
    # ok inside TYPE_CHECKING block
    Alias7 = Optional[Callable[[int], None]]


def func1() -> Optional[Callable[[int], None]]:  # [broken-collections-callable]
    ...

def func2() -> Optional["Callable[[int], None]"]:
    ...

def func3() -> Union[collections.abc.Callable[[int], None], None]:  # [broken-collections-callable]
    ...



# pylint: disable=missing-docstring,too-few-public-methods,expression-not-assigned
class MetaContainer(type):
    __contains__ = None


class NamedExpressionClass(metaclass=MetaContainer):
    if (__iter__ := lambda x: x):  # [unnecessary-lambda-assignment]
        pass


def test():
    1 in NamedExpressionClass()



"""Regression test for alternative Union syntax in runtime contexts.
Syntax support was added in Python 3.10.

The code snipped should not raise any errors.
https://github.com/pylint-dev/pylint/issues/8119
"""
# pylint: disable=missing-docstring,too-few-public-methods
from typing import Generic, TypeVar

T = TypeVar("T")


class Coordinator(Generic[T]):
    def __init__(self, update_interval=None) -> None:
        self.update_interval = update_interval


class Child(Coordinator[int | str]):
    def __init__(self) -> None:
        Coordinator.__init__(self, update_interval=2)

    def _async_update_data(self):
        assert self.update_interval
        self.update_interval = 1



"""Test cases for invalid-name for TypeAlias and related classes with default settings."""
from typing import TypeAlias, Union

# PascalCase names
GoodName: TypeAlias = int
_GoodName: TypeAlias = int
__GoodName: TypeAlias = int
AnotherGoodName = Union[int, str]
GOODName: TypeAlias = int
GOODNAMEType: TypeAlias = int
TodoType: TypeAlias = int
Good2Name: TypeAlias = int
GoodName2: TypeAlias = int

# Non-PascalCase names
BadNAME: TypeAlias = int  # [invalid-name]
badName: TypeAlias = int  # [invalid-name]
AlsoBADName: TypeAlias = int  # [invalid-name]
TBadName: TypeAlias = int  # [invalid-name]
TypeTodo: TypeAlias = int  # [invalid-name]
BadNameT: TypeAlias = int  # [invalid-name]
BAD_NAME = Union[int, str]  # [invalid-name]
_BAD_NAME = Union[int, str]  # [invalid-name]
__BAD_NAME = Union[int, str]  # [invalid-name]
_1BadName = Union[int, str]  # [invalid-name]
ANOTHERBADNAME = Union[int, str]  # [invalid-name]

# Regression tests
# They are not TypeAlias, and thus shouldn't flag the message
x: Union[str, int] = 42
y: Union[str, int]
# But the following, using a good TypeAlias name, is:
GoodTypeAliasToUnion: TypeAlias = Union[str, int]


def my_function():
    """My doc."""
    LocalGoodName: TypeAlias = int
    local_bad_name: TypeAlias = int  # [invalid-name]
    local_declaration: Union[str, int]
    LocalTypeAliasToUnion: TypeAlias = Union[str, int]
    local_declaration = 1
    del local_declaration
    del LocalGoodName, local_bad_name, LocalTypeAliasToUnion



"""Fixture for testing missing documentation in docparams."""
# pylint: disable=broad-exception-raised

def _private_func1(  # [missing-return-doc, missing-return-type-doc, missing-any-param-doc]
    param1,
):
    """This is a test docstring without returns"""
    return param1


def _private_func2(  # [missing-yield-doc, missing-yield-type-doc, missing-any-param-doc]
    param1,
):
    """This is a test docstring without yields"""
    yield param1


def _private_func3(param1):  # [missing-raises-doc, missing-any-param-doc]
    """This is a test docstring without raises"""
    raise Exception("Example")


def public_func1(param1):  # [missing-any-param-doc]
    """This is a test docstring without params"""
    print(param1)


# pylint: disable-next=line-too-long
async def _async_private_func1(  # [missing-return-doc, missing-return-type-doc, missing-any-param-doc]
    param1,
):
    """This is a test docstring without returns"""
    return param1


# pylint: disable-next=line-too-long
async def _async_private_func2(  # [missing-yield-doc, missing-yield-type-doc, missing-any-param-doc]
    param1,
):
    """This is a test docstring without yields"""
    yield param1


async def _async_private_func3(param1):  # [missing-raises-doc, missing-any-param-doc]
    """This is a test docstring without raises"""
    raise Exception("Example")


async def async_public_func1(param1):  # [missing-any-param-doc]
    """This is a test docstring without params"""
    print(param1)


def differing_param_doc(par1: int) -> int:  # [differing-param-doc]
    """This is a test docstring documenting one non-existing param

    :param par1: some param
    :param param: some param
    :return: the sum of the params
    """

    return par1


def differing_param_doc_kwords_only(*, par1: int) -> int:  # [differing-param-doc]
    """This is a test docstring documenting one non-existing param

    :param par1: some param
    :param param: some param
    :return: the sum of the params
    """

    return par1


def missing_type_doc(par1) -> int:  # [missing-type-doc]
    """This is a test docstring params where the type is not specified

    :param par1: some param
    :return: the param
    """

    return par1


def missing_type_doc_kwords_only(*, par1) -> int:  # [missing-type-doc]
    """This is a test docstring params where the type is not specified

    :param par1: some param
    :return: the param
    """

    return par1


def params_are_documented(par1: int, *, par2: int) -> int:
    """This is a test docstring params where nothing is raised as it is all documented

    :param par1: some param
    :param par2: some other param
    :return: the sum of params
    """

    return par1 + par2


# Only check raise nodes within FunctionDefs
raise Exception()



"""This used to raise a non-parent-init-called on Pylint 1.3
See issue https://bitbucket.org/logilab/pylint/issue/308/ for reference"""
# pylint: disable=abstract-method, unused-argument

from tracemalloc import Sequence


class _Traces(Sequence):
    def __init__(self, traces):
        Sequence.__init__(self)



"""Tests for the disabling of bad-option-value."""
# pylint: disable=invalid-name

# pylint: disable=bad-option-value

var = 1  # pylint: disable=a-removed-option

# pylint: enable=bad-option-value

var = 1  # pylint: disable=a-removed-option # [unknown-option-value]

# bad-option-value needs to be disabled before the bad option
var = 1  # pylint: disable=a-removed-option, bad-option-value # [unknown-option-value]
var = 1  # pylint: disable=bad-option-value, a-removed-option



"""Test the generated-members config option."""
# pylint: disable=pointless-statement, invalid-name
from __future__ import annotations
from astroid import nodes
from pylint import checkers

class Klass:
    """A class with a generated member."""

print(Klass().DoesNotExist)
print(Klass().aBC_set1)
print(Klass().ham.does.not_.exist)
print(Klass().spam.does.not_.exist)  # [no-member]
nodes.Tuple.does.not_.exist
checkers.base.doesnotexist()

session = Klass()
SESSION = Klass()
session.rollback()
SESSION.rollback()


# https://github.com/pylint-dev/pylint/issues/6594
# Don't emit no-member inside type annotations
# with PEP 563 'from __future__ import annotations'
print(Klass.X)  # [no-member]
var: "Klass.X"
var2: Klass.X



"""Test case for typevar-name-missing-variance with non-default settings"""
from typing import TypeVar
import typing_extensions as te

# Name set by regex pattern
TypeVarsShouldBeLikeThis = TypeVar("TypeVarsShouldBeLikeThis")
TypeVarsShouldBeLikeThis_contra = TypeVar(
    "TypeVarsShouldBeLikeThis_contra", contravariant=True
)
TypeVarsShouldBeLikeThis_co = TypeVar("TypeVarsShouldBeLikeThis_co", covariant=True)

# Name using the standard style
GoodNameT = TypeVar("GoodNameT")  # [invalid-name]
GoodNameT_co = TypeVar("GoodNameT_co", covariant=True)  # [invalid-name]
GoodNameT_contra = TypeVar("GoodNameT_contra", contravariant=True)  # [invalid-name]


# -- typing_extensions.TypeVar --
TypeVarsShouldBeLikeThis = te.TypeVar("TypeVarsShouldBeLikeThis")
GoodNameT = te.TypeVar("GoodNameT")  # [invalid-name]
GoodNameT_co = te.TypeVar("GoodNameT_co", covariant=True)  # [invalid-name]



"""check unused import from a wildcard import"""
# pylint: disable=line-too-long
from .unused_argument_py3 import *  # [unused-wildcard-import, wildcard-import]
from .unused_global_variable1 import * # [unused-wildcard-import, wildcard-import]
from .unused_import_py30 import * # [unused-wildcard-import, wildcard-import]



# pylint: disable=missing-docstring,invalid-name,line-too-long,too-few-public-methods,use-list-literal,use-dict-literal, typevar-name-incorrect-variance
import typing
import collections
from typing import Generic, TypeVar


# tests/functional/a/assigning_non_slot.py
TYPE = TypeVar('TYPE')

class Cls(Generic[TYPE]):
    """ Simple class with slots """
    __slots__ = ['value']

    def __init__(self, value):
        self.value = value


# tests/functional/d/dangerous_default_value_py30.py
def function4(value=set()): # [dangerous-default-value]
    """set is mutable and dangerous."""
    return value

def function5(value=frozenset()):
    """frozenset is immutable and safe."""
    return value

def function7(value=dict()): # [dangerous-default-value]
    """dict is mutable and dangerous."""
    return value

def function8(value=list()):  # [dangerous-default-value]
    """list is mutable and dangerous."""
    return value

def function17(value=collections.deque()):  # [dangerous-default-value]
    """mutable, dangerous"""
    return value

def function18(value=collections.ChainMap()):  # [dangerous-default-value]
    """mutable, dangerous"""
    return value

def function19(value=collections.Counter()):  # [dangerous-default-value]
    """mutable, dangerous"""
    return value

def function20(value=collections.OrderedDict()):  # [dangerous-default-value]
    """mutable, dangerous"""
    return value

def function21(value=collections.defaultdict()):  # [dangerous-default-value]
    """mutable, dangerous"""
    return value


# tests/functional/p/protocol_classes.py  (min py38)
T2 = typing.TypeVar("T2")

class HasherGeneric(typing.Protocol[T2]):
    """A hashing algorithm, e.g. :func:`hashlib.sha256`."""
    def update(self, blob: bytes):
        ...
    def digest(self) -> bytes:
        ...


# tests/functional/r/regression/regression_2443_duplicate_bases.py
IN = TypeVar('IN', contravariant=True)
OUT = TypeVar('OUT', covariant=True)

class ConsumingMixin(Generic[IN]):
    pass

class ProducingMixin(Generic[OUT]):
    pass

class StreamingMixin(Generic[IN, OUT], ConsumingMixin[IN], ProducingMixin[OUT]):
    pass



"""If the else block returns, it is generally safe to rely on assignments in the except."""
# pylint: disable=missing-function-docstring, invalid-name
import sys

def valid():
    """https://github.com/pylint-dev/pylint/issues/6790"""
    try:
        pass
    except ValueError:
        error = True
    else:
        return

    print(error)


def invalid():
    """The finally will execute before the else returns."""
    try:
        pass
    except ValueError:
        error = None
    else:
        return
    finally:
        print(error)  # [used-before-assignment]


def invalid_2():
    """The else does not return in every branch."""
    try:
        pass
    except ValueError:
        error = None
    else:
        if range(0):
            return
    finally:
        print(error)  # [used-before-assignment]


def invalid_3():
    """Not every except defines the name."""
    try:
        pass
    except ValueError:
        error = None
    except KeyError:
        pass
    finally:
        print(error)  # [used-before-assignment]


def invalid_4():
    """Should not rely on the name in the else even if it returns."""
    try:
        pass
    except ValueError:
        error = True
    else:
        print(error)  # [used-before-assignment]
        return

def valid_exit():
    try:
        pass
    except SystemExit as e:
        lint_result = e.code
    else:
        sys.exit("Bad")
    if lint_result != 0:
        sys.exit("Error is 0.")

    print(lint_result)



"""Tests for no-member on imported modules"""
# pylint: disable=import-outside-toplevel, pointless-statement, missing-function-docstring
# pylint: disable=deprecated-module


def test_no_member_in_getattr():
    """Make sure that a module attribute access is checked by pylint."""
    import math

    math.THIS_does_not_EXIST  # [no-member]


def test_no_member_in_getattr_ignored() -> None:
    """Make sure that a module attribute access check is omitted with a
    module that is configured to be ignored.
    """
    import argparse

    argparse.THIS_does_not_EXIST


def test_ignored_modules_invalid_pattern() -> None:
    import xml

    xml.etree.THIS_does_not_EXIST  # [no-member]


def test_ignored_modules_root_one_applies_as_well() -> None:
    """Check that when a root module is completely ignored, submodules are skipped."""
    import argparse

    argparse.submodule.THIS_does_not_EXIST


def test_ignored_modules_patterns() -> None:
    import collections

    collections.abc.THIS_does_not_EXIST


def test_ignored_classes_no_recursive_pattern() -> None:
    import sys

    sys.THIS_does_not_EXIST  # [no-member]


def test_ignored_classes_qualified_name() -> None:
    """Test that ignored-classes supports qualified name for ignoring."""

    import optparse

    optparse.Values.THIS_does_not_EXIST


def test_ignored_classes_only_name() -> None:
    """Test that ignored_classes works with the name only."""
    import optparse

    optparse.Option.THIS_does_not_EXIST



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Tests for the functional test framework."""

from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.outcomes import Skipped

from pylint import testutils
from pylint.testutils.functional import (
    FunctionalTestFile,
    get_functional_test_files_from_directory,
)

HERE = Path(__file__).parent
DATA_DIRECTORY = HERE / "data"


@pytest.fixture(name="pytest_config")
def pytest_config_fixture() -> MagicMock:
    def _mock_getoption(option: str) -> bool:
        if option == "minimal_messages_config":
            return True
        return False

    config = MagicMock()
    config.getoption.side_effect = _mock_getoption
    return config


def test_parsing_of_pylintrc_init_hook() -> None:
    """Test that we correctly parse an init-hook in a settings file."""
    with pytest.raises(RuntimeError):
        test_file = FunctionalTestFile(str(DATA_DIRECTORY), "init_hook.py")
        testutils.LintModuleTest(test_file)


def test_get_functional_test_files_from_directory() -> None:
    """Test that we correctly check the functional test directory structures."""
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(DATA_DIRECTORY / "u")
    assert exc_info.match("'use_dir.py' should go in 'use'")
    assert exc_info.match(
        "using_dir.py should go in a directory that starts with the "
        "first letters of 'using_dir'"
    )
    assert "incredibly_bold_mischief.py" not in str(exc_info.value)
    # Leading underscore mean that this should not fail the assertion
    get_functional_test_files_from_directory(DATA_DIRECTORY / "u/_no_issue_here")


def test_get_functional_test_files_from_crowded_directory() -> None:
    """Test that we correctly check the functional test directory structures."""
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(
            DATA_DIRECTORY / "m", max_file_per_directory=1
        )
    assert exc_info.match("m: 4 when the max is 1")
    assert exc_info.match("max_overflow: 3 when the max is 1")
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(
            DATA_DIRECTORY / "m", max_file_per_directory=3
        )
    assert exc_info.match("m: 4 when the max is 3")
    assert "max_overflow" not in str(exc_info.value)


def test_minimal_messages_config_enabled(pytest_config: MagicMock) -> None:
    """Test that all messages not targeted in the functional test are disabled
    when running with --minimal-messages-config.
    """
    test_file = FunctionalTestFile(
        str(DATA_DIRECTORY / "m"), "minimal_messages_config.py"
    )
    mod_test = testutils.LintModuleTest(test_file, pytest_config)
    assert all(
        mod_test._linter.is_message_enabled(msgid)
        for msgid in (
            "consider-using-with",
            "unspecified-encoding",
            "consider-using-f-string",
            # Always enable fatal errors: important not to have false negatives
            "astroid-error",
            "fatal",
            "syntax-error",
        )
    )
    assert not mod_test._linter.is_message_enabled("unused-import")


def test_minimal_messages_config_excluded_file(pytest_config: MagicMock) -> None:
    """Test that functional test files can be excluded from the run with
    --minimal-messages-config if they set the exclude_from_minimal_messages_config
    option in their rcfile.
    """
    test_file = FunctionalTestFile(
        str(DATA_DIRECTORY / "m"), "minimal_messages_excluded.py"
    )
    mod_test = testutils.LintModuleTest(test_file, pytest_config)
    with pytest.raises(Skipped):
        mod_test.setUp()



"""Tests for logging-format-interpolation with logging-format-style=new"""

import logging

logging.error("constant string")
logging.error("{}")
logging.error("{}", 1)
logging.error("{0}", 1)
logging.error("{named}", {"named": 1})
logging.error("{} {named}", 1, {"named": 1})
logging.error("{0} {named}", 1, {"named": 1})



# pylint: disable=missing-function-docstring, missing-module-docstring, invalid-name, import-outside-toplevel
# pylint: disable=missing-class-docstring, too-few-public-methods, unused-variable, multiple-statements, line-too-long
"""
Previously, open was uninferable on PyPy so we moved all functional tests
to a separate file. This is no longer the case but the files remain split.
"""
from contextlib import contextmanager
from pathlib import Path

myfile = open("test.txt", encoding="utf-8")  # [consider-using-with]


def test_open():
    fh = open("test.txt", encoding="utf-8")  # [consider-using-with]
    fh.close()

    with open("test.txt", encoding="utf-8") as fh:  # must not trigger
        fh.read()


def test_open_in_enter():
    """Message must not trigger if the resource is allocated in a context manager."""

    class MyContextManager:
        def __init__(self):
            self.file_handle = None

        def __enter__(self):
            self.file_handle = open("foo.txt", "w", encoding="utf-8")  # must not trigger

        def __exit__(self, exc_type, exc_value, traceback):
            self.file_handle.close()


@contextmanager
def test_open_in_with_contextlib():
    """Message must not trigger if the resource is allocated in a context manager."""
    file_handle = open("foo.txt", "w", encoding="utf-8")  # must not trigger
    yield file_handle
    file_handle.close()


def test_open_outside_assignment():
    open("foo", encoding="utf-8").read()  # [consider-using-with]
    content = open("foo", encoding="utf-8").read()  # [consider-using-with]


def test_open_inside_with_block():
    with open("foo", encoding="utf-8") as fh:
        open("bar", encoding="utf-8")  # [consider-using-with]


def test_ternary_if_in_with_block(file1, file2, which):
    """Regression test for issue #4676 (false positive)"""
    with (open(file1, encoding="utf-8") if which else open(file2, encoding="utf-8")) as input_file:  # must not trigger
        return input_file.read()


def test_single_line_with(file1):
    with open(file1, encoding="utf-8"): return file1.read()  # must not trigger


def test_multiline_with_items(file1, file2, which):
    with (open(file1, encoding="utf-8") if which
        else open(file2, encoding="utf-8")) as input_file: return input_file.read()


def test_suppress_on_return():
    return open("foo", encoding="utf8")  # must not trigger


class TestControlFlow:
    """
    The message is triggered if a context manager is assigned to a variable, which name is later
    reassigned without the variable being used inside a ``with`` first.
    E.g. the following would trigger the message:

        a = open("foo")  # <-- would trigger here
        a = "something new"

    But it must not happen that the logic which checks if the same variable is assigned multiple
    times in different code branches where only one of those assign statements is hit at runtime.
    For example, the variable could be assigned in an if-else construct.

    These tests check that the message is not triggered in those circumstances.
    """

    def test_defined_in_if_and_else(self, predicate):
        if predicate:
            file_handle = open("foo", encoding="utf8")  # must not trigger
        else:
            file_handle = open("bar", encoding="utf8")  # must not trigger
        with file_handle:
            return file_handle.read()

    def test_defined_in_else_only(self, predicate):
        if predicate:
            result = "shiny watermelon"
        else:
            file_handle = open("foo", encoding="utf8")  # must not trigger
            with file_handle:
                result = file_handle.read()
        return result

    def test_defined_in_if_only(self, predicate):
        if predicate:
            file_handle = open("foo", encoding="utf8")  # must not trigger
            with file_handle:
                result = file_handle.read()
        else:
            result = "shiny watermelon"
        return result

    def test_triggers_if_reassigned_after_if_else(self, predicate):
        if predicate:
            file_handle = open("foo", encoding="utf8")
        else:
            file_handle = open(  # [consider-using-with]
                "bar", encoding="utf8"
            )
        file_handle = None
        return file_handle

    def test_defined_in_try_and_except(self):
        try:
            file_handle = open("foo", encoding="utf8")  # must not trigger
        except FileNotFoundError:
            file_handle = open("bar", encoding="utf8")  # must not trigger
        with file_handle:
            return file_handle.read()

    def test_defined_in_try_and_finally(self):
        try:
            file_handle = open("foo", encoding="utf8")  # must not trigger
        except FileNotFoundError:
            Path("foo").touch()
        finally:
            # +1: [used-before-assignment]
            file_handle.open("foo", encoding="utf")  # must not trigger consider-using-with
        with file_handle:
            return file_handle.read()

    def test_defined_in_different_except_handlers(self, a, b):
        try:
            result = a/b
        except ZeroDivisionError:
            logfile = open("math_errors.txt", encoding="utf8")  # must not trigger
            result = "Can't divide by zero"
        except TypeError:
            logfile = open("type_errors.txt", encoding="utf8")  # must not trigger
            result = "Wrong types"
        else:
            logfile = open("results.txt", encoding="utf8")  # must not trigger
        with logfile:
            logfile.write(result)

    def test_multiple_return_statements(self, predicate):
        if predicate:
            return open("foo", encoding="utf8")  # must not trigger
        return open("bar", encoding="utf8")  # must not trigger



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Unit tests for pylint.pyreverse.printer_factory."""

from __future__ import annotations

import pytest

from pylint.pyreverse import printer_factory
from pylint.pyreverse.dot_printer import DotPrinter
from pylint.pyreverse.plantuml_printer import PlantUmlPrinter
from pylint.pyreverse.printer import Printer


@pytest.mark.parametrize(
    "filetype, expected_printer_class",
    [
        ("dot", DotPrinter),
        ("puml", PlantUmlPrinter),
        ("plantuml", PlantUmlPrinter),
        ("png", DotPrinter),
    ],
)
def test_get_printer_for_filetype(
    filetype: str, expected_printer_class: type[Printer]
) -> None:
    assert printer_factory.get_printer_for_filetype(filetype) == expected_printer_class



# pylint: disable=missing-docstring,unused-import,import-error, consider-using-from-import
# pylint: disable=wrong-import-order

from pathlib import Path
from some_other_lib import CustomPath as Path   # [shadowed-import]

from pathlib import Path  # [reimported]
import FastAPI.Path as Path  # [shadowed-import]

from pandas._libs import algos
import pandas.core.algorithms as algos  # [shadowed-import]

from sklearn._libs import second as libalgos
import sklearn.core.algorithms as second

import Hello
from goodbye import CustomHello as Hello   # [shadowed-import]



# OPEN BUG: https://github.com/pylint-dev/pylint/issues/8671

class A:
    p: int | None

    def f(self,
          x: (str | None) | (list[A] | list[int]),
          y: A | (int | str) | None,
    ) -> int | str | list[A | int]:
        pass



"""Tests for unnecessary-list-index-lookup."""

# pylint: disable=missing-docstring, too-few-public-methods, expression-not-assigned, line-too-long, unused-variable

my_list = ['a', 'b']

for idx, val in enumerate(my_list):
    print(my_list[idx]) # [unnecessary-list-index-lookup]

for idx, _ in enumerate(my_list):
    print(my_list[0])
    if idx > 0:
        print(my_list[idx - 1])

for idx, val in enumerate(my_list):
    del my_list[idx]

for idx, val in enumerate(my_list):
    my_list[idx] = 42

for vals in enumerate(my_list):
    # This could be refactored, but too complex to infer
    print(my_list[vals[0]])

def process_list(data):
    for index, value in enumerate(data):
        index = 1
        print(data[index])

def process_list_again(data):
    for index, value in enumerate(data):
        value = 1
        print(data[index]) # Can't use value here, it's been redefined

other_list = [1, 2]
for idx, val in enumerate(my_list):
    print(other_list[idx])

OTHER_INDEX = 0
for idx, val in enumerate(my_list):
    print(my_list[OTHER_INDEX])

result = [val for idx, val in enumerate(my_list) if my_list[idx] == 'a'] # [unnecessary-list-index-lookup]
result = [val for idx, val in enumerate(my_list) if idx > 0 and my_list[idx - 1] == 'a']
result = [val for idx, val in enumerate(my_list) if other_list[idx] == 'a']
result = [my_list[idx] for idx, val in enumerate(my_list)] # [unnecessary-list-index-lookup]

# Regression test for https://github.com/pylint-dev/pylint/issues/6049
pairs = [(0, 0)]
for i, (a, b) in enumerate(pairs):
    print(pairs[i][0])

# Regression test for https://github.com/pylint-dev/pylint/issues/6603
for i, num in enumerate():  # raises TypeError, but shouldn't crash pylint
    pass

# Regression test for https://github.com/pylint-dev/pylint/issues/6788
num_list = [1, 2, 3]
for a, b in enumerate(num_list):
    num_list[a], _ = (2, 1)

num_list = [1, 2, 3]
for a, b in enumerate(num_list):
    ([x, num_list[a]], _) = ([5, 6], 1)

# Regression test for https://github.com/pylint-dev/pylint/issues/6818
updated_list = [1, 2, 3]
for idx, val in enumerate(updated_list):
    while updated_list[idx] > 0:
        updated_list[idx] -= 1

updated_list = [1, 2, 3]
for idx, val in enumerate(updated_list):
    print(updated_list[idx]) # [unnecessary-list-index-lookup]
    updated_list[idx] -= 1
    print(updated_list[idx])

# Regression test for https://github.com/pylint-dev/pylint/issues/6896
parts = ["a", "b", "c", "d"]
for i, part in enumerate(parts):
    if i == 3:  # more complex condition actually
        parts.insert(i, "X")
    print(part, parts[i])

# regression tests for https://github.com/pylint-dev/pylint/issues/7682
series = [1, 2, 3, 4, 5]
output_list = [
    (item, series[index])
    for index, item in enumerate(series, start=1)
    if index < len(series)
]

output_list = [
    (item, series[index])
    for index, item in enumerate(series, 1)
    if index < len(series)
]

for idx, val in enumerate(series, start=2):
    print(series[idx])

for idx, val in enumerate(series, 2):
    print(series[idx])

for idx, val in enumerate(series, start=-2):
    print(series[idx])

for idx, val in enumerate(series, -2):
    print(series[idx])

for idx, val in enumerate(series, start=0):
    print(series[idx])  # [unnecessary-list-index-lookup]

for idx, val in enumerate(series, 0):
    print(series[idx])  # [unnecessary-list-index-lookup]

START = 0
for idx, val in enumerate(series, start=START):
    print(series[idx])  # [unnecessary-list-index-lookup]

for idx, val in enumerate(series, START):
    print(series[idx])  # [unnecessary-list-index-lookup]

START = [1, 2, 3]
for i, k in enumerate(series, len(START)):
    print(series[idx])

def return_start(start):
    return start

for i, k in enumerate(series, return_start(20)):
    print(series[idx])

for idx, val in enumerate(iterable=series, start=0):
    print(series[idx])  # [unnecessary-list-index-lookup]

result = [my_list[idx] for idx, val in enumerate(iterable=my_list)]  # [unnecessary-list-index-lookup]

iterable_kwarg = {"iterable": my_list}
result = [my_list[idx] for idx, val in enumerate(**iterable_kwarg)]  # [unnecessary-list-index-lookup]

for idx, val in enumerate():
    print(my_list[idx])

class Command:
    def _get_extra_attrs(self, extra_columns):
        self.extra_rows_start = 8  # pylint: disable=attribute-defined-outside-init
        for index, column in enumerate(extra_columns, start=self.extra_rows_start):
            pass

Y_START = 2
nums = list(range(20))
for y, x in enumerate(nums, start=Y_START + 1):
    pass

for idx, val in enumerate(my_list):
    if (val := 42) and my_list[idx] == 'b':
        print(1)



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

# pylint: disable=too-many-public-methods

from __future__ import annotations

import configparser
import contextlib
import json
import os
import re
import subprocess
import sys
import tempfile
import textwrap
import warnings
from collections.abc import Iterator
from copy import copy
from io import BytesIO, StringIO
from os.path import abspath, dirname, join
from pathlib import Path
from typing import TYPE_CHECKING, Any, TextIO
from unittest import mock
from unittest.mock import patch

import pytest

from pylint import extensions, modify_sys_path
from pylint.constants import MAIN_CHECKER_NAME, MSG_TYPES_STATUS
from pylint.lint.pylinter import PyLinter
from pylint.message import Message
from pylint.reporters import BaseReporter
from pylint.reporters.json_reporter import JSON2Reporter
from pylint.reporters.text import ColorizedTextReporter, TextReporter
from pylint.testutils._run import _add_rcfile_default_pylintrc
from pylint.testutils._run import _Run as Run
from pylint.testutils.utils import (
    _patch_streams,
    _test_cwd,
    _test_environ_pythonpath,
    _test_sys_path,
)
from pylint.utils import utils

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if TYPE_CHECKING:
    from pylint.reporters.ureports.nodes import Section

HERE = abspath(dirname(__file__))
CLEAN_PATH = re.escape(dirname(dirname(__file__)) + os.path.sep)
UNNECESSARY_LAMBDA = join(
    HERE, "functional", "u", "unnecessary", "unnecessary_lambda.py"
)


@contextlib.contextmanager
def _configure_lc_ctype(lc_ctype: str) -> Iterator[None]:
    lc_ctype_env = "LC_CTYPE"
    original_lctype = os.environ.get(lc_ctype_env)
    os.environ[lc_ctype_env] = lc_ctype
    try:
        yield
    finally:
        os.environ.pop(lc_ctype_env)
        if original_lctype:
            os.environ[lc_ctype_env] = original_lctype


class MultiReporter(BaseReporter):
    def __init__(self, reporters: list[BaseReporter]) -> None:
        # pylint: disable=super-init-not-called
        # We don't call it because there is an attribute "linter" that is set inside the base class,
        # and we have another setter here using yet undefined attribute.
        # I don't think fixing the init order in a test class used once is worth it.
        self._reporters = reporters
        self.path_strip_prefix = os.getcwd() + os.sep

    def on_set_current_module(self, *args: str, **kwargs: Any) -> None:
        for rep in self._reporters:
            rep.on_set_current_module(*args, **kwargs)

    def handle_message(self, msg: Message) -> None:
        for rep in self._reporters:
            rep.handle_message(msg)

    def _display(self, layout: Section) -> None:
        pass

    @property
    def out(self) -> TextIO:  # type: ignore[override]
        return self._reporters[0].out

    @property
    def linter(self) -> PyLinter:
        return self._linter

    @linter.setter
    def linter(self, value: PyLinter) -> None:
        self._linter = value
        for rep in self._reporters:
            rep.linter = value


class TestRunTC:
    def _runtest(
        self,
        args: list[str],
        reporter: Any = None,
        out: StringIO | None = None,
        code: int | None = None,
    ) -> None:
        if out is None:
            out = StringIO()
        args = _add_rcfile_default_pylintrc(args)
        pylint_code = self._run_pylint(args, reporter=reporter, out=out)
        if reporter:
            output = reporter.out.getvalue()
        elif hasattr(out, "getvalue"):
            output = out.getvalue()
        else:
            output = None
        msg = f"expected output status {code}, got {pylint_code}"
        if output is not None:
            msg = f"{msg}. Below pylint output: \n{output}"
        assert pylint_code == code, msg

    @staticmethod
    def _run_pylint(args: list[str], out: TextIO, reporter: Any = None) -> int:
        args = _add_rcfile_default_pylintrc([*args, "--persistent=no"])
        with _patch_streams(out):
            with pytest.raises(SystemExit) as cm:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    Run(args, reporter=reporter)
            return int(cm.value.code)

    @staticmethod
    def _clean_paths(output: str) -> str:
        """Normalize path to the tests directory."""
        output = re.sub(CLEAN_PATH, "", output, flags=re.MULTILINE)
        return output.replace("\\", "/")

    def _test_output(
        self, args: list[str], expected_output: str, unexpected_output: str = ""
    ) -> None:
        out = StringIO()
        args = _add_rcfile_default_pylintrc(args)
        self._run_pylint(args, out=out)
        actual_output = self._clean_paths(out.getvalue())
        expected_output = self._clean_paths(expected_output)
        assert expected_output.strip() in actual_output.strip()

        if unexpected_output:
            assert unexpected_output.strip() not in actual_output.strip()

    def _test_output_file(
        self, args: list[str], filename: Path, expected_output: str
    ) -> None:
        """Run Pylint with the ``output`` option set (must be included in
        the ``args`` passed to this method!) and check the file content afterwards.
        """
        out = StringIO()
        args = _add_rcfile_default_pylintrc(args)
        self._run_pylint(args, out=out)
        cmdline_output = out.getvalue()
        file_output = self._clean_paths(Path(filename).read_text(encoding="utf-8"))
        expected_output = self._clean_paths(expected_output)
        assert (
            cmdline_output == ""
        ), "Unexpected output to stdout/stderr while output option was set"
        assert expected_output.strip() in file_output.strip()

    def test_pkginfo(self) -> None:
        """Make pylint check 'pylint.__pkginfo__.py'."""
        # Disable invalid-name because of invalid argument names
        args = ["pylint.__pkginfo__", "--disable=invalid-name"]
        self._runtest(args, reporter=TextReporter(StringIO()), code=0)

    def test_all(self) -> None:
        """Make pylint check itself."""
        reporters = [
            TextReporter(StringIO()),
            ColorizedTextReporter(StringIO()),
            JSON2Reporter(StringIO()),
        ]
        self._runtest(
            [join(HERE, "functional", "a", "arguments.py")],
            reporter=MultiReporter(reporters),
            code=2,
        )

    def test_no_ext_file(self) -> None:
        self._runtest([join(HERE, "input", "noext")], code=0)

    def test_w0704_ignored(self) -> None:
        self._runtest([join(HERE, "input", "ignore_except_pass_by_default.py")], code=0)

    def test_exit_zero(self) -> None:
        self._runtest(
            ["--exit-zero", join(HERE, "regrtest_data", "syntax_error.py")], code=0
        )

    def test_nonexistent_config_file(self) -> None:
        self._runtest(["--rcfile=/tmp/this_file_does_not_exist"], code=32)

    def test_error_missing_arguments(self) -> None:
        self._runtest([], code=32)

    def test_disable_all(self) -> None:
        out = StringIO()
        self._runtest([UNNECESSARY_LAMBDA, "--disable=all"], out=out, code=32)
        assert "No files to lint: exiting." in out.getvalue().strip()

    def test_output_with_verbose(self) -> None:
        out = StringIO()
        self._runtest([UNNECESSARY_LAMBDA, "--verbose"], out=out, code=4)
        assert "Checked 1 files, skipped 0 files" in out.getvalue().strip()

    def test_no_out_encoding(self) -> None:
        """Test redirection of stdout with non ascii characters."""
        # This test reproduces bug #48066 ; it happens when stdout is redirected
        # through '>' : the sys.stdout.encoding becomes then None, and if the
        # output contains non ascii, pylint will crash
        strio = StringIO()
        assert strio.encoding is None
        self._runtest(
            [join(HERE, "regrtest_data", "no_stdout_encoding.py"), "--enable=all"],
            out=strio,
            code=28,
        )

    def test_parallel_execution(self) -> None:
        out = StringIO()
        self._runtest(
            [
                "-j 2",
                join(HERE, "functional", "a", "arguments.py"),
            ],
            out=out,
            code=MSG_TYPES_STATUS["E"],
        )
        assert (
            "Unexpected keyword argument 'fourth' in function call"
            in out.getvalue().strip()
        )

    def test_parallel_execution_missing_arguments(self) -> None:
        self._runtest(["-j 2", "not_here", "not_here_too"], code=1)

    # TODO: PY3.7: Turn off abbreviations in ArgumentsManager after 3.7 support has been dropped
    # argparse changed behaviour with abbreviations on/off in 3.8+ so we can't
    @pytest.mark.xfail
    def test_abbreviations_are_not_supported(self) -> None:
        expected = "No module named --load-plugin"
        self._test_output([".", "--load-plugin"], expected_output=expected)

    def test_enable_all_works(self) -> None:
        module = join(HERE, "data", "clientmodule_test.py")
        expected = textwrap.dedent(
            f"""
        ************* Module data.clientmodule_test
        {module}:9:8: W0612: Unused variable 'local_variable' (unused-variable)
        {module}:17:4: C0116: Missing function or method docstring (missing-function-docstring)
        {module}:21:0: C0115: Missing class docstring (missing-class-docstring)
        """
        )
        self._test_output(
            [module, "--disable=I", "--enable=all", "-rn"], expected_output=expected
        )

    def test_wrong_import_position_when_others_disabled(self) -> None:
        module1 = join(HERE, "regrtest_data", "import_something.py")
        module2 = join(HERE, "regrtest_data", "wrong_import_position.py")
        expected_output = textwrap.dedent(
            f"""
        ************* Module wrong_import_position
        {module2}:11:0: C0413: Import "import os" should be placed at the top of the module (wrong-import-position)
        """
        )
        args = [
            module2,
            module1,
            "--disable=all",
            "--enable=wrong-import-position",
            "-rn",
            "-sn",
        ]
        out = StringIO()
        self._run_pylint(args, out=out)
        actual_output = self._clean_paths(out.getvalue().strip())

        to_remove = "No config file found, using default configuration"
        if to_remove in actual_output:
            actual_output = actual_output[len(to_remove) :]
        if actual_output.startswith("Using config file "):
            # If ~/.pylintrc is present remove the
            # Using config file...  line
            actual_output = actual_output[actual_output.find("\n") :]
        assert self._clean_paths(expected_output.strip()) == actual_output.strip()

    def test_type_annotation_names(self) -> None:
        """Test resetting the `_type_annotation_names` list to `[]` when leaving a module.

        An import inside `module_a`, which is used as a type annotation in `module_a`, should not prevent
        emitting the `unused-import` message when the same import occurs in `module_b` & is unused.
        See: https://github.com/pylint-dev/pylint/issues/4150
        """
        module1 = join(
            HERE, "regrtest_data", "imported_module_in_typehint", "module_a.py"
        )

        module2 = join(
            HERE, "regrtest_data", "imported_module_in_typehint", "module_b.py"
        )
        expected_output = textwrap.dedent(
            f"""
        ************* Module module_b
        {module2}:1:0: W0611: Unused import uuid (unused-import)
        """
        )
        args = [
            module1,
            module2,
            "--disable=all",
            "--enable=unused-import",
            "-rn",
            "-sn",
        ]
        out = StringIO()
        self._run_pylint(args, out=out)
        actual_output = self._clean_paths(out.getvalue().strip())
        assert self._clean_paths(expected_output.strip()) in actual_output.strip()

    def test_import_itself_not_accounted_for_relative_imports(self) -> None:
        expected = "Your code has been rated at 10.00/10"
        package = join(HERE, "regrtest_data", "dummy")
        self._test_output(
            [package, "--disable=locally-disabled", "-rn"], expected_output=expected
        )

    def test_reject_empty_indent_strings(self) -> None:
        expected = "Option cannot be an empty string"
        module = join(HERE, "data", "clientmodule_test.py")
        self._test_output([module, "--indent-string="], expected_output=expected)

    def test_json_report_when_file_has_syntax_error(self) -> None:
        out = StringIO()
        module = join(HERE, "regrtest_data", "syntax_error.py")
        self._runtest([module], code=2, reporter=JSON2Reporter(out))
        output = json.loads(out.getvalue())["messages"]
        assert isinstance(output, list)
        assert len(output) == 1
        assert isinstance(output[0], dict)
        expected = {
            "obj": "",
            "column": 9,
            "line": 1,
            "type": "error",
            "symbol": "syntax-error",
            "module": "syntax_error",
        }
        message = output[0]
        for key, value in expected.items():
            assert key in message
            assert message[key] == value
        msg = message["message"].lower()
        assert any(x in msg for x in ("expected ':'", "invalid syntax"))
        assert "syntax_error" in msg
        assert "line 1" in msg

    def test_json_report_when_file_is_missing(self) -> None:
        out = StringIO()
        module = join(HERE, "regrtest_data", "totally_missing.py")
        self._runtest([module], code=1, reporter=JSON2Reporter(out))
        output = json.loads(out.getvalue())["messages"]
        assert isinstance(output, list)
        assert len(output) == 1
        assert isinstance(output[0], dict)
        expected = {
            "obj": "",
            "column": 0,
            "line": 1,
            "type": "fatal",
            "symbol": "fatal",
            "module": module,
        }
        message = output[0]
        for key, value in expected.items():
            assert key in message
            assert message[key] == value
        assert message["message"].startswith("No module named")

    def test_json_report_does_not_escape_quotes(self) -> None:
        out = StringIO()
        module = join(HERE, "regrtest_data", "unused_variable.py")
        self._runtest([module], code=4, reporter=JSON2Reporter(out))
        output = json.loads(out.getvalue())["messages"]
        assert isinstance(output, list)
        assert len(output) == 1
        assert isinstance(output[0], dict)
        expected = {
            "symbol": "unused-variable",
            "module": "unused_variable",
            "column": 4,
            "message": "Unused variable 'variable'",
            "messageId": "W0612",
            "line": 4,
            "type": "warning",
        }
        message = output[0]
        for key, value in expected.items():
            assert key in message
            assert message[key] == value

    def test_information_category_disabled_by_default(self) -> None:
        expected = "Your code has been rated at 10.00/10"
        path = join(HERE, "regrtest_data", "meta.py")
        self._test_output([path], expected_output=expected)

    def test_error_mode_shows_no_score(self) -> None:
        module = join(HERE, "regrtest_data", "application_crash.py")
        expected_output = textwrap.dedent(
            f"""
        ************* Module application_crash
        {module}:1:6: E0602: Undefined variable 'something_undefined' (undefined-variable)
        """
        )
        self._test_output([module, "-E"], expected_output=expected_output)

    def test_evaluation_score_shown_by_default(self) -> None:
        expected_output = "Your code has been rated at "
        module = join(HERE, "regrtest_data", "application_crash.py")
        self._test_output([module], expected_output=expected_output)

    def test_confidence_levels(self) -> None:
        expected = "Your code has been rated at"
        path = join(HERE, "regrtest_data", "meta.py")
        self._test_output(
            [path, "--confidence=HIGH,INFERENCE"], expected_output=expected
        )

    def test_bom_marker(self) -> None:
        path = join(HERE, "regrtest_data", "meta.py")
        expected = "Your code has been rated at 10.00/10"
        self._test_output([path, "-rn"], expected_output=expected)

    def test_pylintrc_plugin_duplicate_options(self) -> None:
        dummy_plugin_path = join(HERE, "regrtest_data", "dummy_plugin")
        # Enable --load-plugins=dummy_plugin
        sys.path.append(dummy_plugin_path)
        config_path = join(HERE, "regrtest_data", "dummy_plugin.rc")
        expected = (
            ":dummy-message-01 (I9061): *Dummy short desc 01*\n"
            "  Dummy long desc This message belongs to the dummy_plugin checker.\n\n"
            ":dummy-message-02 (I9060): *Dummy short desc 02*\n"
            "  Dummy long desc This message belongs to the dummy_plugin checker."
        )
        self._test_output(
            [
                f"--rcfile={config_path}",
                "--help-msg",
                "dummy-message-01",
                "dummy-message-02",
            ],
            expected_output=expected,
        )
        expected = (
            "[DUMMY_PLUGIN]\n\n# Dummy option 1\ndummy_option_1=dummy value 1\n\n"
            "# Dummy option 2\ndummy_option_2=dummy value 2"
        )
        self._test_output(
            [f"--rcfile={config_path}", "--generate-rcfile"], expected_output=expected
        )
        sys.path.remove(dummy_plugin_path)

    def test_pylintrc_comments_in_values(self) -> None:
        path = join(HERE, "regrtest_data", "test_pylintrc_comments.py")
        config_path = join(HERE, "regrtest_data", "comments_pylintrc")
        expected = textwrap.dedent(
            f"""
        ************* Module test_pylintrc_comments
        {path}:2:0: W0311: Bad indentation. Found 1 spaces, expected 4 (bad-indentation)
        {path}:1:0: C0114: Missing module docstring (missing-module-docstring)
        {path}:1:0: C0116: Missing function or method docstring (missing-function-docstring)
        """
        )
        self._test_output(
            [path, f"--rcfile={config_path}", "-rn"], expected_output=expected
        )

    def test_no_crash_with_formatting_regex_defaults(self) -> None:
        self._runtest(
            ["--ignore-patterns=a"], reporter=TextReporter(StringIO()), code=32
        )

    def test_getdefaultencoding_crashes_with_lc_ctype_utf8(self) -> None:
        module = join(HERE, "regrtest_data", "application_crash.py")
        expected_output = textwrap.dedent(
            f"""
        {module}:1:6: E0602: Undefined variable 'something_undefined' (undefined-variable)
        """
        )
        with _configure_lc_ctype("UTF-8"):
            self._test_output([module, "-E"], expected_output=expected_output)

    @pytest.mark.skipif(sys.platform == "win32", reason="only occurs on *nix")
    def test_parseable_file_path(self) -> None:
        file_name = "test_target.py"
        fake_path = HERE + os.getcwd()
        module = join(fake_path, file_name)

        try:
            # create module under directories which have the same name as reporter.path_strip_prefix
            # e.g. /src/some/path/src/test_target.py when reporter.path_strip_prefix = /src/
            os.makedirs(fake_path)
            with open(module, "w", encoding="utf-8") as test_target:
                test_target.write("a,b = object()")

            self._test_output(
                [module, "--output-format=parseable"], expected_output=file_name
            )
        finally:
            os.remove(module)
            os.removedirs(fake_path)

    @pytest.mark.parametrize(
        "input_path,module,expected_path",
        [
            (join(HERE, "mymodule.py"), "mymodule", join(HERE, "mymodule.py")),
            ("mymodule.py", "mymodule", "mymodule.py"),
        ],
    )
    def test_stdin(self, input_path: str, module: str, expected_path: str) -> None:
        expected_output = f"""************* Module {module}
{expected_path}:1:0: W0611: Unused import os (unused-import)

"""

        with mock.patch(
            "pylint.lint.pylinter._read_stdin", return_value="import os\n"
        ) as mock_stdin:
            self._test_output(
                ["--from-stdin", input_path, "--disable=all", "--enable=unused-import"],
                expected_output=expected_output,
            )
            assert mock_stdin.call_count == 1

    def test_stdin_missing_modulename(self) -> None:
        self._runtest(["--from-stdin"], code=32)

    @pytest.mark.parametrize("write_bpy_to_disk", [False, True])
    def test_relative_imports(self, write_bpy_to_disk: bool, tmp_path: Path) -> None:
        a = tmp_path / "a"

        b_code = textwrap.dedent(
            """
            from .c import foobar
            from .d import bla  # module does not exist

            foobar('hello')
            bla()
            """
        )

        c_code = textwrap.dedent(
            """
            def foobar(arg):
                pass
            """
        )

        a.mkdir()
        (a / "__init__.py").write_text("")
        if write_bpy_to_disk:
            (a / "b.py").write_text(b_code)
        (a / "c.py").write_text(c_code)

        with _test_cwd(tmp_path):
            # why don't we start pylint in a sub-process?
            expected = (
                "************* Module a.b\n"
                "a/b.py:3:0: E0401: Unable to import 'a.d' (import-error)\n\n"
            )

            if write_bpy_to_disk:
                # --from-stdin is not used here
                self._test_output(
                    ["a/b.py", "--disable=all", "--enable=import-error"],
                    expected_output=expected,
                )

            # this code needs to work w/ and w/o a file named a/b.py on the hard disk.
            with mock.patch("pylint.lint.pylinter._read_stdin", return_value=b_code):
                self._test_output(
                    [
                        "--from-stdin",
                        join("a", "b.py"),
                        "--disable=all",
                        "--enable=import-error",
                    ],
                    expected_output=expected,
                )

    def test_stdin_syntax_error(self) -> None:
        expected_output = """************* Module a
a.py:1:4: E0001: Parsing failed: 'invalid syntax (a, line 1)' (syntax-error)"""
        with mock.patch(
            "pylint.lint.pylinter._read_stdin", return_value="for\n"
        ) as mock_stdin:
            self._test_output(
                ["--from-stdin", "a.py", "--disable=all", "--enable=syntax-error"],
                expected_output=expected_output,
            )
            assert mock_stdin.call_count == 1

    def test_version(self) -> None:
        def check(lines: list[str]) -> None:
            assert lines[0].startswith("pylint ")
            assert lines[1].startswith("astroid ")
            assert lines[2].startswith("Python ")

        out = StringIO()
        self._run_pylint(["--version"], out=out)
        check(out.getvalue().splitlines())

        result = subprocess.check_output([sys.executable, "-m", "pylint", "--version"])
        result_str = result.decode("utf-8")
        check(result_str.splitlines())

    def test_fail_under(self) -> None:
        self._runtest(
            [
                "--fail-under",
                "-10",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_plus7_5.py"),
            ],
            code=0,
        )
        self._runtest(
            [
                "--fail-under",
                "6",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_plus7_5.py"),
            ],
            code=0,
        )
        self._runtest(
            [
                "--fail-under",
                "7.5",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_plus7_5.py"),
            ],
            code=0,
        )
        self._runtest(
            [
                "--fail-under",
                "7.6",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_plus7_5.py"),
            ],
            code=16,
        )

        self._runtest(
            [
                "--fail-under",
                "-11",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_minus10.py"),
            ],
            code=0,
        )
        self._runtest(
            [
                "--fail-under",
                "-10",
                "--enable=all",
                join(HERE, "regrtest_data", "fail_under_minus10.py"),
            ],
            code=0,
        )
        # Need the old evaluation formula to test a negative score
        # failing below a negative --fail-under threshold
        self._runtest(
            [
                "--fail-under",
                "-9",
                "--enable=all",
                "--evaluation",
                "0 if fatal else 10.0 - ((float(5 * error + warning + refactor + convention) / statement) * 10)",
                join(HERE, "regrtest_data", "fail_under_minus10.py"),
            ],
            code=22,
        )
        self._runtest(
            [
                "--fail-under",
                "-5",
                "--enable=all",
                "--evaluation",
                "0 if fatal else 10.0 - ((float(5 * error + warning + refactor + convention) / statement) * 10)",
                join(HERE, "regrtest_data", "fail_under_minus10.py"),
            ],
            code=22,
        )

    @pytest.mark.parametrize(
        "fu_score,fo_msgs,fname,out",
        [
            # Essentially same test cases as --fail-under, but run with/without a detected
            # issue code missing-function-docstring (C0116) is issue in both files
            # --fail-under should be irrelevant as missing-function-docstring is hit
            (-10, "missing-function-docstring", "fail_under_plus7_5.py", 16),
            (6, "missing-function-docstring", "fail_under_plus7_5.py", 16),
            (7.5, "missing-function-docstring", "fail_under_plus7_5.py", 16),
            (7.6, "missing-function-docstring", "fail_under_plus7_5.py", 16),
            (-11, "missing-function-docstring", "fail_under_minus10.py", 22),
            (-10, "missing-function-docstring", "fail_under_minus10.py", 22),
            (-9, "missing-function-docstring", "fail_under_minus10.py", 22),
            (-5, "missing-function-docstring", "fail_under_minus10.py", 22),
            # --fail-under should guide whether error code as missing-function-docstring is not hit
            (-10, "broad-exception-caught", "fail_under_plus7_5.py", 0),
            (6, "broad-exception-caught", "fail_under_plus7_5.py", 0),
            (7.5, "broad-exception-caught", "fail_under_plus7_5.py", 0),
            (7.6, "broad-exception-caught", "fail_under_plus7_5.py", 16),
            (-11, "broad-exception-caught", "fail_under_minus10.py", 0),
            (-10, "broad-exception-caught", "fail_under_minus10.py", 0),
            (-9, "broad-exception-caught", "fail_under_minus10.py", 22),
            (-5, "broad-exception-caught", "fail_under_minus10.py", 22),
            # Enable by message id
            (-10, "C0116", "fail_under_plus7_5.py", 16),
            # Enable by category
            (-10, "C", "fail_under_plus7_5.py", 16),
            (-10, "fake1,C,fake2", "fail_under_plus7_5.py", 16),
            # Ensure entire category not enabled by any msg id
            (-10, "C0115", "fail_under_plus7_5.py", 0),
        ],
    )
    def test_fail_on(self, fu_score: int, fo_msgs: str, fname: str, out: int) -> None:
        self._runtest(
            [
                "--fail-under",
                f"{fu_score:f}",
                f"--fail-on={fo_msgs}",
                "--enable=all",
                join(HERE, "regrtest_data", fname),
                # Use the old form of the evaluation that can go negative
                "--evaluation",
                "0 if fatal else 10.0 - ((float(5 * error + warning + refactor + convention) / statement) * 10)",
            ],
            code=out,
        )

    @pytest.mark.parametrize(
        "opts,out",
        [
            # Special case to ensure that disabled items from category aren't enabled
            (["--disable=C0116", "--fail-on=C"], 0),
            # Ensure order does not matter
            (["--fail-on=C", "--disable=C0116"], 0),
            # Ensure --fail-on takes precedence over --disable
            (["--disable=C0116", "--fail-on=C0116"], 16),
            # Ensure order does not matter
            (["--fail-on=C0116", "--disable=C0116"], 16),
            # Message emitted by PyLinter itself
            (
                [
                    "--fail-on=unknown-option-value",
                    "--disable=all",
                    "--enable=unknown-option-value, trigger",
                ],
                4,
            ),
        ],
    )
    def test_fail_on_edge_case(self, opts: list[str], out: int) -> None:
        self._runtest(
            [*opts, join(HERE, "regrtest_data", "fail_under_plus7_5.py")],
            code=out,
        )

    @staticmethod
    def test_modify_sys_path() -> None:
        # pylint: disable = too-many-statements
        cwd = "/tmp/pytest-of-root/pytest-0/test_do_not_import_files_from_0"
        default_paths = [
            "/usr/local/lib/python39.zip",
            "/usr/local/lib/python3.9",
            "/usr/local/lib/python3.9/lib-dynload",
            "/usr/local/lib/python3.9/site-packages",
        ]
        with _test_sys_path(), patch("os.getcwd") as mock_getcwd:
            mock_getcwd.return_value = cwd
            paths = [cwd, *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath():
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = ["", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath():
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = [".", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath():
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = ["/do_not_remove", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath():
                modify_sys_path()
            assert sys.path == paths

            paths = [cwd, cwd, *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath("."):
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = [cwd, "/custom_pythonpath", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath("/custom_pythonpath"):
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = [cwd, "/custom_pythonpath", cwd, *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath("/custom_pythonpath:"):
                modify_sys_path()
            assert sys.path == [paths[1]] + paths[3:]

            paths = ["", cwd, "/custom_pythonpath", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath(":/custom_pythonpath"):
                modify_sys_path()
            assert sys.path == paths[2:]

            paths = [cwd, cwd, "/custom_pythonpath", *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath(":/custom_pythonpath:"):
                modify_sys_path()
            assert sys.path == paths[2:]

            paths = [cwd, cwd, *default_paths]
            sys.path = copy(paths)
            with _test_environ_pythonpath(":."):
                modify_sys_path()
            assert sys.path == paths[1:]
            sys.path = copy(paths)
            with _test_environ_pythonpath(f":{cwd}"):
                modify_sys_path()
            assert sys.path == paths[1:]

            sys.path = copy(paths)
            with _test_environ_pythonpath(".:"):
                modify_sys_path()
            assert sys.path == paths[1:]
            sys.path = copy(paths)
            with _test_environ_pythonpath(f"{cwd}:"):
                modify_sys_path()
            assert sys.path == paths[1:]

            paths = ["", cwd, *default_paths, cwd]
            sys.path = copy(paths)
            with _test_environ_pythonpath(cwd):
                modify_sys_path()
            assert sys.path == paths[1:]

    @staticmethod
    def test_plugin_that_imports_from_open() -> None:
        """Test that a plugin that imports a source file from a checker open()
        function (ala pylint_django) does not raise an exception."""
        with _test_sys_path():
            # Enable --load-plugins=importing_plugin
            sys.path.append(join(HERE, "regrtest_data", "importing_plugin"))
            with _test_cwd(join(HERE, "regrtest_data", "settings_project")):
                Run(
                    ["--load-plugins=importing_plugin", "models.py"],
                    exit=False,
                )

    @pytest.mark.parametrize(
        "args",
        [
            ["--disable=import-error,unused-import"],
            # Test with multiple jobs for 'hmac.py' for which we have a
            # CVE against: https://github.com/pylint-dev/pylint/issues/959
            ["-j2", "--disable=import-error,unused-import"],
        ],
    )
    def test_do_not_import_files_from_local_directory(
        self, tmp_path: Path, args: list[str]
    ) -> None:
        for path in ("astroid.py", "hmac.py"):
            file_path = tmp_path / path
            file_path.write_text("'Docstring'\nimport completely_unknown\n")
            pylint_call = [sys.executable, "-m", "pylint", *args, path]
            with _test_cwd(tmp_path):
                subprocess.check_output(pylint_call, cwd=str(tmp_path))
            new_python_path = os.environ.get("PYTHONPATH", "").strip(":")
            with _test_cwd(tmp_path), _test_environ_pythonpath(f"{new_python_path}:"):
                # Appending a colon to PYTHONPATH should not break path stripping
                # https://github.com/pylint-dev/pylint/issues/3636
                subprocess.check_output(pylint_call, cwd=str(tmp_path))

    @staticmethod
    def test_import_plugin_from_local_directory_if_pythonpath_cwd(
        tmp_path: Path,
    ) -> None:
        p_plugin = tmp_path / "plugin.py"
        p_plugin.write_text("# Some plugin content")
        if sys.platform == "win32":
            python_path = "."
        else:
            python_path = f"{os.environ.get('PYTHONPATH', '').strip(':')}:."
        with _test_cwd(tmp_path), _test_environ_pythonpath(python_path):
            args = [sys.executable, "-m", "pylint", "--load-plugins", "plugin"]
            process = subprocess.run(
                args, cwd=str(tmp_path), stderr=subprocess.PIPE, check=False
            )
            assert (
                "AttributeError: module 'plugin' has no attribute 'register'"
                in process.stderr.decode()
            )

    def test_allow_import_of_files_found_in_modules_during_parallel_check(
        self, tmp_path: Path
    ) -> None:
        test_directory = tmp_path / "test_directory"
        test_directory.mkdir()
        spam_module = test_directory / "spam.py"
        spam_module.write_text("'Empty'")

        init_module = test_directory / "__init__.py"
        init_module.write_text("'Empty'")

        # For multiple jobs we could not find the `spam.py` file.
        with _test_cwd(tmp_path):
            args = [
                "-j2",
                "--disable=missing-docstring, missing-final-newline",
                "test_directory",
            ]
            self._runtest(args, code=0)

        # A single job should be fine as well
        with _test_cwd(tmp_path):
            args = [
                "-j1",
                "--disable=missing-docstring, missing-final-newline",
                "test_directory",
            ]
            self._runtest(args, code=0)

    @staticmethod
    def test_can_list_directories_without_dunder_init(tmp_path: Path) -> None:
        test_directory = tmp_path / "test_directory"
        test_directory.mkdir()
        spam_module = test_directory / "spam.py"
        spam_module.write_text("'Empty'")

        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "pylint",
                "--disable=missing-docstring, missing-final-newline",
                "test_directory",
            ],
            cwd=str(tmp_path),
            stderr=subprocess.PIPE,
        )

    @pytest.mark.needs_two_cores
    def test_jobs_score(self) -> None:
        path = join(HERE, "regrtest_data", "unused_variable.py")
        expected = "Your code has been rated at 7.50/10"
        self._test_output([path, "--jobs=2", "-ry"], expected_output=expected)

    def test_regression_parallel_mode_without_filepath(self) -> None:
        # Test that parallel mode properly passes filepath
        # https://github.com/pylint-dev/pylint/issues/3564
        path = join(
            HERE, "regrtest_data", "regression_missing_init_3564", "subdirectory/"
        )
        self._test_output([path, "-j2"], expected_output="")

    def test_output_file_valid_path(self, tmp_path: Path) -> None:
        path = join(HERE, "regrtest_data", "unused_variable.py")
        output_file = tmp_path / "output.txt"
        expected = "Your code has been rated at 7.50/10"
        self._test_output_file(
            [path, f"--output={output_file}"],
            output_file,
            expected_output=expected,
        )

    def test_output_file_invalid_path_exits_with_code_32(self) -> None:
        path = join(HERE, "regrtest_data", "unused_variable.py")
        output_file = "thisdirectorydoesnotexit/output.txt"
        self._runtest([path, f"--output={output_file}"], code=32)

    @pytest.mark.parametrize(
        "args, expected",
        [
            ([], 0),
            (["--enable=C"], 0),
            (["--fail-on=superfluous-parens"], 0),
            (["--fail-on=import-error"], 6),
            (["--fail-on=unused-import"], 6),
            (["--fail-on=unused-import", "--enable=C"], 22),
            (["--fail-on=missing-function-docstring"], 22),
            (["--fail-on=useless-suppression"], 6),
            (["--fail-on=useless-suppression", "--enable=C"], 22),
        ],
    )
    def test_fail_on_exit_code(self, args: list[str], expected: int) -> None:
        path = join(HERE, "regrtest_data", "fail_on.py")
        # We set fail-under to be something very low so that even with the warnings
        # and errors that are generated they don't affect the exit code.
        self._runtest([path, "--fail-under=-10", "--disable=C", *args], code=expected)

    def test_one_module_fatal_error(self) -> None:
        """Fatal errors in one of several modules linted still exits non-zero."""
        valid_path = join(HERE, "conftest.py")
        invalid_path = join(HERE, "garbagePath.py")
        self._runtest([valid_path, invalid_path, "--disable=C"], code=1)

    @pytest.mark.parametrize(
        "args, expected",
        [
            ([], 0),
            (["--enable=C"], 0),
            (["--fail-on=superfluous-parens"], 0),
            (["--fail-on=import-error"], 0),
            (["--fail-on=unused-import"], 0),
            (["--fail-on=unused-import", "--enable=C"], 0),
            (["--fail-on=missing-function-docstring"], 0),
            (["--fail-on=useless-suppression"], 1),
            (["--fail-on=useless-suppression", "--enable=C"], 1),
        ],
    )
    def test_fail_on_info_only_exit_code(self, args: list[str], expected: int) -> None:
        path = join(HERE, "regrtest_data", "fail_on_info_only.py")
        self._runtest([path, *args], code=expected)

    @pytest.mark.parametrize(
        "output_format, expected_output",
        [
            (
                "text",
                "{path}:4:4: W0612: Unused variable 'variable' (unused-variable)",
            ),
            (
                "parseable",
                "{path}:4: [W0612(unused-variable), test] Unused variable 'variable'",
            ),
            (
                "msvs",
                "{path}(4): [W0612(unused-variable)test] Unused variable 'variable'",
            ),
            (
                "colorized",
                (
                    "{path}:4:4: W0612: \x1B[35mUnused variable 'variable'\x1B[0m (\x1B[35munused-variable\x1B[0m)"
                ),
            ),
            ("json", '"message": "Unused variable \'variable\'",'),
            ("json2", '"message": "Unused variable \'variable\'",'),
        ],
    )
    def test_output_file_can_be_combined_with_output_format_option(
        self, tmp_path: Path, output_format: str, expected_output: str
    ) -> None:
        path = join(HERE, "regrtest_data", "unused_variable.py")
        output_file = tmp_path / "output.txt"
        self._test_output_file(
            [path, f"--output={output_file}", f"--output-format={output_format}"],
            output_file,
            expected_output.format(path="tests/regrtest_data/unused_variable.py"),
        )

    def test_output_file_can_be_combined_with_custom_reporter(
        self, tmp_path: Path
    ) -> None:
        path = join(HERE, "regrtest_data", "unused_variable.py")
        output_file = tmp_path / "output.txt"
        # It does not really have to be a truly custom reporter.
        # It is only important that it is being passed explicitly to ``Run``.
        myreporter = TextReporter()
        self._run_pylint(
            [path, f"--output={output_file}"],
            out=sys.stdout,
            reporter=myreporter,
        )
        assert output_file.exists()

    def test_output_file_specified_in_rcfile(self, tmp_path: Path) -> None:
        output_file = tmp_path / "output.txt"
        rcfile = tmp_path / "pylintrc"
        rcfile_contents = textwrap.dedent(
            f"""
        [MAIN]
        output={output_file}
        """
        )
        rcfile.write_text(rcfile_contents, encoding="utf-8")
        path = join(HERE, "regrtest_data", "unused_variable.py")
        expected = "Your code has been rated at 7.50/10"
        self._test_output_file(
            [path, f"--output={output_file}", f"--rcfile={rcfile}"],
            output_file,
            expected_output=expected,
        )

    @staticmethod
    def test_load_text_repoter_if_not_provided() -> None:
        """Test if PyLinter.reporter is a TextReporter if no reporter is provided."""
        linter = PyLinter()

        assert isinstance(linter.reporter, TextReporter)

    @staticmethod
    def test_regex_paths_csv_validator() -> None:
        """Test to see if _regexp_paths_csv_validator works.
        Previously the validator crashed when encountering already validated values.
        Reported in https://github.com/pylint-dev/pylint/issues/5437
        """
        with pytest.raises(SystemExit) as ex:
            args = _add_rcfile_default_pylintrc(
                ["--ignore-paths", "test", join(HERE, "regrtest_data", "empty.py")]
            )
            Run(args)
        assert ex.value.code == 0

    @staticmethod
    def test_max_inferred_for_complicated_class_hierarchy() -> None:
        """Regression test for a crash reported in https://github.com/pylint-dev/pylint/issues/5679.

        The class hierarchy of 'sqlalchemy' is so intricate that it becomes uninferable with
        the standard max_inferred of 100. We used to crash when this happened.
        """
        with pytest.raises(SystemExit) as ex:
            path = join(
                HERE, "regrtest_data", "max_inferable_limit_for_classes", "main.py"
            )
            Run([path])
        # Error code should not include bit-value 1 for crash
        assert not ex.value.code % 2

    def test_recursive(self) -> None:
        """Tests if running linter over directory using --recursive=y"""
        self._runtest(
            [join(HERE, "regrtest_data", "directory", "subdirectory"), "--recursive=y"],
            code=0,
        )

    def test_recursive_globbing(self) -> None:
        """Tests if running linter over directory using --recursive=y and globbing"""
        self._runtest(
            [join(HERE, "regrtest_data", "d?rectory", "subd*"), "--recursive=y"],
            code=0,
        )

    @pytest.mark.parametrize("ignore_value", ["ignored_subdirectory", "failing.py"])
    def test_ignore_recursive(self, ignore_value: str) -> None:
        """Tests recursive run of linter ignoring directory using --ignore parameter.

        Ignored directory contains files yielding lint errors. If directory is not ignored
        test would fail due these errors.
        """
        directory = join(HERE, "regrtest_data", "directory")
        self._runtest([directory, "--recursive=y", f"--ignore={ignore_value}"], code=0)

    @pytest.mark.parametrize("ignore_pattern_value", ["ignored_.*", "failing.*"])
    def test_ignore_pattern_recursive(self, ignore_pattern_value: str) -> None:
        """Tests recursive run of linter ignoring directory using --ignore-parameter parameter.

        Ignored directory contains files yielding lint errors. If directory is not ignored
        test would fail due these errors.
        """
        directory = join(HERE, "regrtest_data", "directory")
        self._runtest(
            [directory, "--recursive=y", f"--ignore-patterns={ignore_pattern_value}"],
            code=0,
        )

    def test_ignore_pattern_from_stdin(self) -> None:
        """Test if linter ignores standard input if the filename matches the ignore pattern."""
        with mock.patch("pylint.lint.pylinter._read_stdin", return_value="import os\n"):
            self._runtest(
                [
                    "--from-stdin",
                    "mymodule.py",
                    "--disable=all",
                    "--enable=unused-import",
                    "--ignore-patterns=mymodule.py",
                ],
                code=0,
            )

    @pytest.mark.parametrize("ignore_path_value", [".*ignored.*", ".*failing.*"])
    def test_ignore_path_recursive(self, ignore_path_value: str) -> None:
        """Tests recursive run of linter ignoring directory using --ignore-path parameter.

        Ignored directory contains files yielding lint errors. If directory is not ignored
        test would fail due these errors.
        """
        directory = join(HERE, "regrtest_data", "directory")
        self._runtest(
            [directory, "--recursive=y", f"--ignore-paths={ignore_path_value}"], code=0
        )

    def test_recursive_current_dir(self) -> None:
        with _test_sys_path():
            # pytest is including directory HERE/regrtest_data to sys.path which causes
            # astroid to believe that directory is a package.
            sys.path = [
                path
                for path in sys.path
                if not os.path.basename(path) == "regrtest_data"
            ]
            with _test_cwd():
                os.chdir(join(HERE, "regrtest_data", "directory", "subdirectory"))
                self._runtest(
                    [".", "--recursive=y"],
                    code=0,
                )

    def test_ignore_path_recursive_current_dir(self) -> None:
        """Tests that path is normalized before checked that is ignored. GitHub issue #6964"""
        with _test_sys_path():
            # pytest is including directory HERE/regrtest_data to sys.path which causes
            # astroid to believe that directory is a package.
            sys.path = [
                path
                for path in sys.path
                if not os.path.basename(path) == "regrtest_data"
            ]
            with _test_cwd():
                os.chdir(join(HERE, "regrtest_data", "directory"))
                self._runtest(
                    [
                        ".",
                        "--recursive=y",
                        "--ignore-paths=^ignored_subdirectory/.*",
                    ],
                    code=0,
                )

    def test_syntax_error_invalid_encoding(self) -> None:
        module = join(HERE, "regrtest_data", "invalid_encoding.py")
        expected_output = "unknown encoding"
        self._test_output([module, "-E"], expected_output=expected_output)

    @pytest.mark.parametrize(
        "module_name,expected_output",
        [
            ("good.py", ""),
            ("bad_wrong_num.py", "(syntax-error)"),
            ("bad_missing_num.py", "(bad-file-encoding)"),
        ],
    )
    def test_encoding(self, module_name: str, expected_output: str) -> None:
        path = join(HERE, "regrtest_data", "encoding", module_name)
        self._test_output(
            [path], expected_output=expected_output, unexpected_output="(astroid-error)"
        )

    def test_line_too_long_useless_suppression(self) -> None:
        """A test that demonstrates a known false positive for useless-suppression

        See https://github.com/pylint-dev/pylint/issues/3368

        If you manage to make this test fail and remove the useless-suppression
        warning please contact open a Pylint PR!
        """
        module = join(HERE, "regrtest_data", "line_too_long_no_code.py")
        expected = textwrap.dedent(
            f"""
        {module}:1:0: I0011: Locally disabling line-too-long (C0301) (locally-disabled)
        {module}:1:0: I0021: Useless suppression of 'line-too-long' (useless-suppression)
        """
        )

        self._test_output([module, "--enable=all"], expected_output=expected)

    def test_output_no_header(self) -> None:
        module = join(HERE, "data", "clientmodule_test.py")
        expected = "Unused variable 'local_variable'"
        not_expected = textwrap.dedent(
            """************* Module data.clientmodule_test"""
        )

        args = [module, "--output-format=no-header"]
        self._test_output(
            args, expected_output=expected, unexpected_output=not_expected
        )

    def test_no_name_in_module(self) -> None:
        """Test that a package with both a variable name `base` and a module `base`
        does not emit a no-name-in-module msg."""
        module = join(HERE, "regrtest_data", "test_no_name_in_module.py")
        unexpected = "No name 'errors' in module 'list' (no-name-in-module)"
        self._test_output(
            [module, "-E"], expected_output="", unexpected_output=unexpected
        )


class TestCallbackOptions:
    """Test for all callback options we support."""

    @staticmethod
    @pytest.mark.parametrize(
        "command,expected",
        [
            (["--list-msgs"], "Emittable messages with current interpreter:"),
            (["--list-msgs-enabled"], "Enabled messages:"),
            (["--list-groups"], "nonascii-checker"),
            (["--list-conf-levels"], "Confidence(name='HIGH', description="),
            (["--list-extensions"], "pylint.extensions.empty_comment"),
            (["--full-documentation"], "Pylint global options and switches"),
            (["--long-help"], "Environment variables:"),
        ],
    )
    def test_output_of_callback_options(
        command: list[str], expected: str, tmp_path: Path
    ) -> None:
        """Test whether certain strings are in the output of a callback command."""
        command = _add_rcfile_default_pylintrc(command)
        process = subprocess.run(
            [sys.executable, "-m", "pylint", *command],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        assert expected in process.stdout

    @staticmethod
    @pytest.mark.parametrize(
        "args,expected,error",
        [
            [["--help-msg", "W0101"], ":unreachable (W0101)", False],
            [["--help-msg", "WX101"], "No such message id", False],
            [["--help-msg"], "--help-msg: expected at least one argumen", True],
            [["--help-msg", "C0102,C0103"], ":invalid-name (C0103):", False],
        ],
    )
    def test_help_msg(
        args: list[str], expected: str, error: bool, tmp_path: Path
    ) -> None:
        """Test the --help-msg flag."""
        args = _add_rcfile_default_pylintrc(args)
        process = subprocess.run(
            [sys.executable, "-m", "pylint", *args],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        if error:
            result = process.stderr
        else:
            result = process.stdout
        assert expected in result

    @staticmethod
    def test_generate_rcfile(tmp_path: Path) -> None:
        """Test the --generate-rcfile flag."""
        args = _add_rcfile_default_pylintrc(["--generate-rcfile"])
        process = subprocess.run(
            [sys.executable, "-m", "pylint", *args],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        assert "[MAIN]" in process.stdout
        assert "[MASTER]" not in process.stdout
        assert "profile" not in process.stdout
        args = _add_rcfile_default_pylintrc(["--generate-rcfile"])
        process_two = subprocess.run(
            [sys.executable, "-m", "pylint", *args],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        assert process.stdout == process_two.stdout

        # Check that the generated file is valid
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp:
            filename = temp.name
            temp.write(process.stdout)
            runner = Run(
                [join(HERE, "regrtest_data", "empty.py"), f"--rcfile={filename}"],
                exit=False,
            )
            assert not runner.linter.msg_status
        os.remove(filename)

    @staticmethod
    def test_generate_config_disable_symbolic_names() -> None:
        """Test that --generate-rcfile puts symbolic names in the --disable option."""
        out = StringIO()
        with _patch_streams(out):
            with pytest.raises(SystemExit):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    Run(["--generate-rcfile", "--rcfile=", "--persistent=no"])
        output = out.getvalue()

        # Get rid of the pesky messages that pylint emits if the
        # configuration file is not found.
        pattern = rf"\[{MAIN_CHECKER_NAME.upper()}"
        main = re.search(pattern, output)
        assert main is not None, f"{pattern} not found in {output}"

        out = StringIO(output[main.start() :])
        parser = configparser.RawConfigParser()
        parser.read_file(out)
        messages = utils._splitstrip(parser.get("MESSAGES CONTROL", "disable"))
        assert "suppressed-message" in messages

    @staticmethod
    def test_generate_toml_config(tmp_path: Path) -> None:
        """Test the --generate-toml-config flag."""
        args = _add_rcfile_default_pylintrc(
            [
                "--preferred-modules=a:b",
                "--generate-toml-config",
            ]
        )
        process = subprocess.run(
            [sys.executable, "-m", "pylint", *args],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        assert "[tool.pylint.main]" in process.stdout
        assert "[tool.pylint.master]" not in process.stdout
        assert '"positional arguments"' not in process.stdout
        assert '"optional arguments"' not in process.stdout
        assert 'preferred-modules = ["a:b"]' in process.stdout

        process_two = subprocess.run(
            [sys.executable, "-m", "pylint", *args],
            capture_output=True,
            encoding="utf-8",
            check=False,
            cwd=str(tmp_path),
        )
        assert process.stdout == process_two.stdout

        # Check that the generated file is valid
        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".toml", delete=False
        ) as temp:
            filename = temp.name
            temp.write(process.stdout)
            runner = Run(
                [join(HERE, "regrtest_data", "empty.py"), f"--rcfile={filename}"],
                exit=False,
            )
            assert not runner.linter.msg_status
        os.remove(filename)

    @staticmethod
    def test_generate_toml_config_disable_symbolic_names() -> None:
        """Test that --generate-toml-config puts symbolic names in the --disable option."""
        output_stream = StringIO()
        with _patch_streams(output_stream):
            with pytest.raises(SystemExit):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    Run(["--generate-toml-config"])

        out = output_stream.getvalue()
        bytes_out = BytesIO(out.encode("utf-8"))
        content = tomllib.load(bytes_out)
        messages = content["tool"]["pylint"]["messages control"]["disable"]
        assert "useless-suppression" in messages, out

    @staticmethod
    def test_errors_only() -> None:
        """Test the --errors-only flag."""
        with pytest.raises(SystemExit):
            run = Run(["--errors-only"])
            assert run.linter._error_mode

    @staticmethod
    def test_errors_only_functions_as_disable() -> None:
        """--errors-only functions as a shortcut for --disable=W,C,R,I;
        it no longer enables any messages."""
        run = Run(
            [str(UNNECESSARY_LAMBDA), "--disable=import-error", "--errors-only"],
            exit=False,
        )
        assert not run.linter.is_message_enabled("import-error")

    @staticmethod
    def test_verbose() -> None:
        """Test the --verbose flag."""
        with pytest.raises(SystemExit):
            run = Run(["--verbose"])
            assert run.verbose

        with pytest.raises(SystemExit):
            run = Run(["--verbose=True"])
            assert run.verbose

    @staticmethod
    def test_enable_all_extensions() -> None:
        """Test to see if --enable-all-extensions does indeed load all extensions."""
        # Record all extensions
        plugins = []
        for filename in os.listdir(os.path.dirname(extensions.__file__)):
            if filename.endswith(".py") and not filename.startswith("_"):
                plugins.append(f"pylint.extensions.{filename[:-3]}")

        # Check if they are loaded
        runner = Run(
            ["--enable-all-extensions", join(HERE, "regrtest_data", "empty.py")],
            exit=False,
        )
        assert sorted(plugins) == sorted(runner.linter._dynamic_plugins)



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Unit tests for the imports checker."""

import os

import astroid
from pytest import CaptureFixture

from pylint.checkers import imports
from pylint.interfaces import UNDEFINED
from pylint.testutils import CheckerTestCase, MessageTest
from pylint.testutils._run import _Run as Run

REGR_DATA = os.path.join(os.path.dirname(__file__), "..", "regrtest_data", "")


class TestImportsChecker(CheckerTestCase):
    CHECKER_CLASS = imports.ImportsChecker

    def test_relative_beyond_top_level(self) -> None:
        module = astroid.MANAGER.ast_from_module_name("beyond_top", REGR_DATA)
        import_from = module.body[0]

        msg = MessageTest(
            msg_id="relative-beyond-top-level",
            node=import_from,
            line=1,
            col_offset=0,
            end_line=1,
            end_col_offset=25,
        )
        with self.assertAddsMessages(msg):
            self.checker.visit_importfrom(import_from)
        with self.assertNoMessages():
            self.checker.visit_importfrom(module.body[1])
        with self.assertNoMessages():
            self.checker.visit_importfrom(module.body[2].body[0])

    @staticmethod
    def test_relative_beyond_top_level_two(capsys: CaptureFixture[str]) -> None:
        Run(
            [
                f"{os.path.join(REGR_DATA, 'beyond_top_two')}",
                "-d all",
                "-e relative-beyond-top-level",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()

        top_level_function = os.path.join(
            REGR_DATA, "beyond_top_two/namespace_package/top_level_function.py"
        )
        Run(
            [top_level_function, "-d all", "-e relative-beyond-top-level"],
            exit=False,
        )
        output2, errors2 = capsys.readouterr()

        assert len(output.split("\n")) == 5
        assert len(output2.split("\n")) == 5
        assert errors == errors2

    @staticmethod
    def test_relative_beyond_top_level_three(capsys: CaptureFixture[str]) -> None:
        Run(
            [
                f"{os.path.join(REGR_DATA, 'beyond_top_three/a.py')}",
                "-d all",
                "-e relative-beyond-top-level",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()
        assert len(output.split("\n")) == 5
        assert errors == ""

    @staticmethod
    def test_relative_beyond_top_level_four(capsys: CaptureFixture[str]) -> None:
        Run(
            [
                f"{os.path.join(REGR_DATA, 'beyond_top_four/module')}",
                "-d missing-docstring,unused-import",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()
        assert len(output.split("\n")) == 5
        assert errors == ""

    def test_wildcard_import_init(self) -> None:
        module = astroid.MANAGER.ast_from_module_name("init_wildcard", REGR_DATA)
        import_from = module.body[0]

        with self.assertNoMessages():
            self.checker.visit_importfrom(import_from)

    def test_wildcard_import_non_init(self) -> None:
        module = astroid.MANAGER.ast_from_module_name("wildcard", REGR_DATA)
        import_from = module.body[0]

        msg = MessageTest(
            msg_id="wildcard-import",
            node=import_from,
            args="empty",
            confidence=UNDEFINED,
            line=1,
            col_offset=0,
            end_line=1,
            end_col_offset=19,
        )
        with self.assertAddsMessages(msg):
            self.checker.visit_importfrom(import_from)

    @staticmethod
    def test_preferred_module(capsys: CaptureFixture[str]) -> None:
        """
        Tests preferred-module configuration option
        """
        # test preferred-modules case with base module import
        Run(
            [
                f"{os.path.join(REGR_DATA, 'preferred_module/unpreferred_module.py')}",
                "-d all",
                "-e preferred-module",
                # prefer sys instead of os (for triggering test)
                "--preferred-modules=os:sys",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()

        # assert that we saw preferred-modules triggered
        assert "Prefer importing 'sys' instead of 'os'" in output
        # assert there were no errors
        assert len(errors) == 0

        # test preferred-modules trigger case with submodules
        Run(
            [
                f"{os.path.join(REGR_DATA, 'preferred_module/unpreferred_submodule.py')}",
                "-d all",
                "-e preferred-module",
                # prefer os.path instead of pathlib (for triggering test)
                "--preferred-modules=os.path:pathlib",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()

        # assert that we saw preferred-modules triggered
        assert "Prefer importing 'pathlib' instead of 'os.path'" in output
        # assert there were no errors
        assert len(errors) == 0

        # test preferred-modules ignore case with submodules
        Run(
            [
                f"{os.path.join(REGR_DATA, 'preferred_module/unpreferred_submodule.py')}",
                "-d all",
                "-e preferred-module",
                # prefer pathlib instead of os.stat (for untriggered test)
                "--preferred-modules=os.stat:pathlib",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()

        # assert that we didn't see preferred-modules triggered
        assert "Your code has been rated at 10.00/10" in output
        # assert there were no errors
        assert len(errors) == 0

        # test preferred-modules base module for implemented submodule (from ... import)
        Run(
            [
                f"{os.path.join(REGR_DATA, 'preferred_module/unpreferred_submodule.py')}",
                "-d all",
                "-e preferred-module",
                # prefer pathlib instead of os (for triggering test)
                "--preferred-modules=os:pathlib",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()

        # assert that we saw preferred-modules triggered with base module
        assert "Prefer importing 'pathlib' instead of 'os'" in output
        # assert there were no errors
        assert len(errors) == 0

        # Test for challenges with preferred modules indefinite matches
        Run(
            [
                f"{os.path.join(REGR_DATA, 'preferred_module/unpreferred_submodule.py')}",
                "-d all",
                "-e preferred-module",
                # prefer pathlib instead of random (testing to avoid regression)
                # pathlib shouldn't match with path, which is in the test file
                "--preferred-modules=random:pathlib",
            ],
            exit=False,
        )
        _, errors = capsys.readouterr()

        # Assert there were no errors
        assert len(errors) == 0

    @staticmethod
    def test_allow_reexport_package(capsys: CaptureFixture[str]) -> None:
        """Test --allow-reexport-from-package option."""

        # Option disabled - useless-import-alias should always be emitted
        Run(
            [
                f"{os.path.join(REGR_DATA, 'allow_reexport')}",
                "--allow-reexport-from-package=no",
                "-sn",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()
        assert len(output.split("\n")) == 7, f"Expected 7 line breaks in:{output}"
        assert (
            "__init__.py:1:0: C0414: Import alias does not rename original package (useless-import-alias)"
            in output
        )
        assert (
            "file.py:2:0: C0414: Import alias does not rename original package (useless-import-alias)"
            in output
        )
        assert len(errors) == 0

        # Option enabled - useless-import-alias should only be emitted for 'file.py'
        Run(
            [
                f"{os.path.join(REGR_DATA, 'allow_reexport')}",
                "--allow-reexport-from-package=yes",
                "--disable=missing-module-docstring",
                "-sn",
            ],
            exit=False,
        )
        output, errors = capsys.readouterr()
        assert len(output.split("\n")) == 3
        assert "__init__.py" not in output
        assert (
            "file.py:2:0: C0414: Import alias does not rename original package (useless-import-alias)"
            in output
        )
        assert len(errors) == 0



"""Test case for typevar-name-incorrect-variance with default settings"""
# pylint: disable=line-too-long

from typing import TypeVar

T_co = TypeVar("T_co", covariant=True, contravariant=True) # [typevar-double-variance,typevar-name-incorrect-variance]
T_co = TypeVar("T_co", covariant=True, contravariant=False)
T_contra = TypeVar("T_contra", covariant=True, contravariant=True) # [typevar-double-variance,typevar-name-incorrect-variance]
T_contra = TypeVar("T_contra", covariant=False, contravariant=True)
T = TypeVar("T", covariant=False, contravariant=False)



"""Regression test for https://github.com/pylint-dev/pylint/issues/6531."""

# pylint: disable=missing-docstring, redefined-outer-name

import pytest


class Wallet:
    def __init__(self):
        self.balance = 0

    def add_cash(self, earned):
        self.balance += earned

    def spend_cash(self, spent):
        self.balance -= spent

@pytest.fixture
def my_wallet():
    '''Returns a Wallet instance with a zero balance'''
    return Wallet()

@pytest.mark.parametrize("earned,spent,expected", [
    (30, 10, 20),
    (20, 2, 18),
])
def test_transactions(my_wallet, earned, spent, expected):
    my_wallet.add_cash(earned)
    my_wallet.spend_cash(spent)
    assert my_wallet.balance == expected



# pylint: disable=invalid-name,missing-docstring,pointless-statement,unnecessary-comprehension

var = (1, 2, 3)

for x in var:
    pass
for x in (1, 2, 3):
    pass
for x in [1, 2, 3]:  # [consider-using-tuple]
    pass

(x for x in var)
(x for x in (1, 2, 3))
(x for x in [1, 2, 3])  # [consider-using-tuple]

[x for x in var]
[x for x in (1, 2, 3)]
[x for x in [1, 2, 3]]  # [consider-using-tuple]


for x in [*var]:  # [consider-using-tuple]
    pass
for x in [2, *var]:  # [consider-using-tuple]
    pass

[x for x in [*var, 2]]  # [consider-using-tuple]


# Don't emit warning for sets as this is handled by builtin checker
(x for x in {1, 2, 3})  # [use-sequence-for-iteration]
[x for x in {*var, 2}]



"""Check for method without self as first argument"""
# pylint: disable=missing-docstring,too-few-public-methods

MYSTATICMETHOD = staticmethod


def returns_staticmethod(my_function):
    """Create a staticmethod from function `my_function`"""
    return staticmethod(my_function)


class NoSelfArgument:
    """dummy class"""

    def __init__(truc):  # [no-self-argument]
        """method without self"""
        print(1)

    def abdc(yoo):  # [no-self-argument]
        """another test"""
        print(yoo)

    def edf(self):
        """just another method"""
        print('yapudju in', self)

    @staticmethod
    def say_hello():
        """A standard staticmethod"""
        print("hello!")

    @MYSTATICMETHOD
    def say_goodbye():
        """A staticmethod but using a different name"""
        print("goodbye!")

    @returns_staticmethod
    def concatenate_strings(string1, string2):
        """A staticmethod created by `returns_staticmethod` function"""
        return string1 + string2

    def varargs(*args):  # [no-self-argument]
        """A method without a self argument but with *args."""

    def kwargs(**kwargs):  # [no-self-argument]
        """A method without a self argument but with **kwargs."""

    def varargs_and_kwargs(*args, **kwargs):  # [no-self-argument]
        """A method without a self argument but with *args and **kwargs."""


class Toto:

    def __class_getitem__(cls, params):
        # This is actually a special method which is always a class method.
        # See https://www.python.org/dev/peps/pep-0560/#class-getitem
        pass

    def __class_other__(cls, params):  # [no-self-argument]
        # This is not a special case and as such is an instance method.
        pass
